from ada import Material
from ada.materials.metals import CarbonSteel


def write_ifc_mat(material: Material):
    if material.parent is None:
        raise ValueError("Parent cannot be None")

    a = material.parent.get_assembly()
    ifc_store = a.ifc_store
    f = ifc_store.f

    ifc_mat = f.create_entity("IfcMaterial", Name=material.name, Category="Steel")

    properties = []
    if type(material.model) is CarbonSteel:
        strength_grade = f.create_entity("IfcText", material.model.grade)
        strength_grade_prop = f.create_entity("IfcPropertySingleValue", Name="Grade", NominalValue=strength_grade)
        properties.append(strength_grade_prop)

    if material.model.sig_y is not None:
        yield_stress = f.create_entity("IfcPressureMeasure", float(material.model.sig_y))
        properties += [
            f.create_entity(
                "IfcPropertySingleValue",
                Name="YieldStress",
                NominalValue=yield_stress,
            )
        ]

    mass_density = f.create_entity("IfcMassDensityMeasure", float(material.model.rho))
    young_modulus = f.create_entity("IfcModulusOfElasticityMeasure", float(material.model.E))
    poisson_ratio = f.create_entity("IfcPositiveRatioMeasure", float(material.model.v))
    therm_exp_coeff = f.create_entity("IfcThermalExpansionCoefficientMeasure", float(material.model.alpha))
    specific_heat = f.create_entity("IfcSpecificHeatCapacityMeasure", float(material.model.zeta))
    properties += [
        f.create_entity(
            "IfcPropertySingleValue",
            Name="YoungModulus",
            NominalValue=young_modulus,
        ),
        f.create_entity(
            "IfcPropertySingleValue",
            Name="PoissonRatio",
            NominalValue=poisson_ratio,
        ),
        f.create_entity(
            "IfcPropertySingleValue",
            Name="ThermalExpansionCoefficient",
            NominalValue=therm_exp_coeff,
        ),
        f.create_entity(
            "IfcPropertySingleValue",
            Name="SpecificHeatCapacity",
            NominalValue=specific_heat,
        ),
        f.create_entity("IfcPropertySingleValue", Name="MassDensity", NominalValue=mass_density),
    ]

    f.create_entity(
        "IfcMaterialProperties",
        Name="MaterialMechanical",
        Description="A Material property description",
        Properties=properties,
        Material=ifc_mat,
    )

    return ifc_mat



from meshio.abaqus._abaqus import abaqus_to_meshio_type as aba_meshio_original

from ada.fem.shapes.definitions import ElemShapeTypes

solids = ElemShapeTypes.solids
shells = ElemShapeTypes.shell
lines = ElemShapeTypes.lines

gmsh_to_meshio_ordering = {
    solids.TETRA10: [0, 1, 2, 3, 4, 5, 6, 7, 9, 8],
    solids.HEX20: [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13, 9, 16, 18, 19, 17, 10, 12, 14, 15],
    solids.HEX27: [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13, 9, 16, 18, 19, 17, 10, 12, 14, 15, 22, 23, 21, 24, 20, 25, 26],
    solids.WEDGE15: [0, 1, 2, 3, 4, 5, 6, 9, 7, 12, 14, 13, 8, 10, 11],
    solids.PYRAMID13: [0, 1, 2, 3, 4, 5, 8, 10, 6, 7, 9, 11, 12],
    lines.LINE3: [0, 2, 1],
}

aba_to_meshio_types = aba_meshio_original

meshio_convert_default = dict(
    hexahedron=solids.HEX8,
    hexahedron20=solids.HEX20,
    triangle=shells.TRI,
    tetra10=solids.TETRA10,
    line=lines.LINE,
)

meshio_to_abaqus_type = {
    v: k if v not in meshio_convert_default.keys() else meshio_convert_default[v]
    for k, v in aba_to_meshio_types.items()
}

meshio_to_ada_type = {}



from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Union

from ada.concepts.points import Node

from .common import FemBase
from .elements import Elem
from .sets import FemSet

if TYPE_CHECKING:
    from ada import FEM


class SurfTypes:
    ELEMENT = "ELEMENT"
    NODE = "NODE"

    all = [ELEMENT, NODE]


@dataclass
class ElemSurface:
    fem_set: FemSet
    side_index: int


class Surface(FemBase):
    """
    Documentation

        https://abaqus-docs.mit.edu/2017/English/SIMACAEKEYRefMap/simakey-r-surface.htm#simakey-r-surface__simakey-r-surface-s-datadesc5


    Parameters.

    :param name: Unique name of surface
    :param surf_type: Type of surface
    :param fem_set:
    :param weight_factor:
    :param id_refs: Explicitly defined by list of tuple [(elid/nid,spos), ..]
    :param parent:
    :param metadata:
    """

    TYPES = SurfTypes

    def __init__(
        self,
        name,
        surf_type,
        fem_set: Union[FemSet, List[FemSet]],
        weight_factor=None,
        el_face_index: Union[int, List[int]] = None,
        id_refs=None,
        parent: "FEM" = None,
        metadata=None,
    ):
        super().__init__(name, metadata, parent)

        self._type = surf_type.upper()

        if self.type not in SurfTypes.all:
            raise ValueError(f'Surface type "{self.type}" is currently not supported\\implemented. Valid types are')

        self._fem_set = fem_set
        if type(fem_set) is list:
            if type(el_face_index) is not list:
                raise ValueError("You cannot define a list of FemSets and not also include a List of el_face_indices")

        self._weight_factor = weight_factor
        self._el_face_index = el_face_index
        self._id_refs = id_refs
        self._refs = []
        if isinstance(fem_set, FemSet):
            fem_set.refs.append(self)

    @property
    def type(self):
        return self._type

    @property
    def fem_set(self) -> Union[FemSet, List[FemSet]]:
        return self._fem_set

    @fem_set.setter
    def fem_set(self, value: Union[FemSet, List[FemSet]]):
        self._fem_set = value

    @property
    def weight_factor(self):
        return self._weight_factor

    @property
    def el_face_index(self) -> Union[int, List[int]]:
        return self._el_face_index

    @property
    def id_refs(self):
        return self._id_refs

    @property
    def refs(self):
        return self._refs


def create_surface_from_nodes(surface_name: str, nodes: List[Node], fem: "FEM", shell_positive=True) -> Surface:
    from ada.fem.elements import find_element_type_from_list
    from ada.fem.shapes import ElemType

    all_el = [el for n in nodes for el in filter(lambda x: type(x) is Elem, n.refs)]
    el_type = find_element_type_from_list(all_el)

    surf_map = {
        ElemType.SOLID: get_surface_from_nodes_on_solid_elements,
        ElemType.SHELL: get_surface_from_nodes_on_shell_elements,
    }
    surf_writer = surf_map.get(el_type, None)

    if surf_writer is None:
        raise NotImplementedError(f'Currently Surface writing on element type "{el_type}" is not supported')

    return surf_writer(surface_name, all_el, nodes, fem, shell_positive)


def get_surface_from_nodes_on_solid_elements(
    surface_name: str, all_el: List[Elem], nodes: List[Node], fem: "FEM", shell_positive: bool
) -> Surface:
    elements = []
    face_seq_indices = {}
    for el in all_el:
        paralell_face_index = elem_has_parallel_face(el, nodes)
        if paralell_face_index is None:
            continue

        if el not in elements:
            face_seq_indices[el] = paralell_face_index
            elements.append(el)

    fsets = []
    fset_el_face_indices = []
    for el, el_face_index in face_seq_indices.items():
        side_name = f"S{el_face_index + 1}"
        fs = FemSet(f"_{surface_name}_{el.id}_{side_name}", [el])
        if fs.name in fem.sets.elements.keys():
            fs_elem_1 = fem.sets.elements[fs.name]
            fs_elem_1.add_members([el])
        else:
            fs_elem_1 = fem.add_set(fs)
        fsets.append(fs_elem_1)
        fset_el_face_indices.append(el_face_index)

    return Surface(surface_name, Surface.TYPES.ELEMENT, fsets, el_face_index=fset_el_face_indices)


def get_surface_from_nodes_on_shell_elements(
    surface_name: str, all_el: List[Elem], nodes: List[Node], fem: "FEM", shell_positive: bool
) -> Surface:
    elements = []
    for el in all_el:
        if elem_has_parallel_face(el, nodes) is None:
            continue
        elements.append(el)

    side_name = 1 if shell_positive is True else -1
    fs = fem.add_set(FemSet(f"_{surface_name}_{side_name}", elements))

    return Surface(surface_name, Surface.TYPES.ELEMENT, fs, el_face_index=side_name)


def elem_has_parallel_face(el: Elem, nodes: List[Node]):
    for i, nid_refs in enumerate(el.shape.faces_seq):
        all_face_nodes_in_plane = True
        for nid in nid_refs:
            no = el.nodes[nid]
            if no not in nodes:
                all_face_nodes_in_plane = False
                break
        if all_face_nodes_in_plane is True:
            return i
    return None



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from ada import Material
from ada.config import get_logger

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly

logger = get_logger()


def materials_str(assembly: Assembly) -> str:
    all_mat = chain.from_iterable([p.materials for p in assembly.get_all_parts_in_assembly(True)])
    all_mat_unique = {x.name: x for x in all_mat}
    nl_geom = False
    for x in assembly.fem.get_all_steps():
        if x.nl_geom is True:
            nl_geom = True

    return "\n".join([material_str(mat, nl_geom) for mat in all_mat_unique.values()])


def material_str(material: Material, nl_geom) -> str:
    from ada.core.utils import NewLine

    # Bi-linear hardening ECRO_LINE=_F(D_SIGM_EPSI=2.0e06, SY=2.35e06,)

    model = material.model
    nl = NewLine(3, suffix="	")
    has_plasticity = model.plasticity_model is not None and model.plasticity_model.eps_p is not None
    if has_plasticity and nl_geom is True:
        nl_mat = "nl_mat=(	\n	"
        eps = [e for e in model.plasticity_model.eps_p]
        eps[0] = 1e-5  # Epsilon index=0 cannot be zero
        nl_mat += "".join([f"{e:.4E},{s:.4E}," + next(nl) for e, s in zip(eps, model.plasticity_model.sig_p)]) + ")"
        nl_mat += """
Traction=DEFI_FONCTION(
    NOM_PARA='EPSI', NOM_RESU='SIGM', VALE=nl_mat, INTERPOL='LIN', PROL_DROITE='LINEAIRE', PROL_GAUCHE='CONSTANT'
)"""
        mat_nl_in = ", TRACTION=_F(SIGM=Traction,)"
    else:
        logger.debug(f"No plasticity is defined for material {material.name}")
        nl_mat = ""
        mat_nl_in = ""

    return f"""{nl_mat}

{material.name} = DEFI_MATERIAU(
    ELAS=_F(E={model.E}, NU={model.v}, RHO={model.rho}){mat_nl_in},
)
"""



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem.exceptions.element_support import IncompatibleElements
from ada.fem.formats.utils import get_fem_model_from_assembly
from ada.fem.loads import LoadGravity
from ada.fem.steps import Step

from .write.write_elements import must_be_converted_to_general_section

if TYPE_CHECKING:
    from ada import Assembly


def check_compatibility(assembly: Assembly):
    p = get_fem_model_from_assembly(assembly)
    step = assembly.fem.steps[0] if len(assembly.fem.steps) > 0 else None

    if step is not None:
        step_types = [x.type for x in assembly.fem.steps]
        if Step.TYPES.EIGEN in step_types:
            for line in p.fem.elements.lines:
                if must_be_converted_to_general_section(line.fem_sec.section.type):
                    raise IncompatibleElements("Calculix does not support general beam elements in Eigenvalue analysis")
        if step.TYPES.STATIC in step_types:
            static_step = assembly.fem.steps[step_types.index(step.TYPES.STATIC)]
            has_gravity = False
            for load in static_step.loads:
                if isinstance(load, LoadGravity):
                    has_gravity = True
            for line in p.fem.elements.lines:
                if must_be_converted_to_general_section(line.fem_sec.section.type) and has_gravity:
                    raise IncompatibleElements("The Calculix general U1 elements does not work with gravity loads")



from __future__ import annotations

import base64
import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from ada.sat.utils import create_sat_from_beams

if TYPE_CHECKING:
    from ada import Part


def embed_sat_geometry(root: ET.Element, part: Part) -> dict:
    geom = ET.SubElement(root, "geometry")
    sat_embedded = ET.SubElement(geom, "sat_embedded", dict(encoding="base64", compression="zip", tag_name="dnvscp"))

    # Encode and compress the string
    sat_map = dict()
    beams_sat = create_sat_from_beams(part)
    sat_map.update(beams_sat.sat_map)

    data = create_sesam_sat_bytes(beams_sat.sat_text)

    encoded_data = base64.b64encode(data).decode()

    # Add the encoded data to the element using CDATA
    sat_embedded.text = "<![CDATA[" + encoded_data + "]]>"

    return sat_map


def create_sesam_sat_bytes(sat_body_str: str) -> bytes:
    from datetime import datetime

    date = datetime.now()
    date_string = date.strftime("%d %a %b %d %H:%M:%S %Y")

    header = f"""2000 0 1 0
18 SESAM - gmGeometry 14 ACIS 30.0.1 NT {date_string}
1000 9.9999999999999995e-07 1e-10\n"""

    footer = "End-of-ACIS-data "

    return bytes(header + sat_body_str + footer, encoding="utf8")



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Iterable, List

import numpy as np

from ada.config import Settings

from .exceptions import VectorNormalizeError


class Plane(Enum):
    XY = "xy"
    XZ = "xz"
    YZ = "yz"


@dataclass
class EquationOfPlane:
    point_in_plane: tuple | list | np.ndarray
    normal: tuple | list | np.ndarray
    yvec: tuple | list | np.ndarray = None

    PLANE: ClassVar[Plane]

    @staticmethod
    def from_arbitrary_points(points):
        points = np.array(points)
        normal = normal_to_points_in_plane(points)
        pip = points[0]
        return EquationOfPlane(pip, normal)

    def __post_init__(self):
        point_in_plane = self.point_in_plane
        normal = self.normal
        x1, y1, z1 = point_in_plane
        a = normal[0]
        b = normal[1]
        c = normal[2]
        self.d = -(a * x1 + b * y1 + c * z1)

    def return_points_in_plane(self, points: np.ndarray) -> np.ndarray:
        return points[points.dot(self.normal) + self.d == 0]

    def is_point_in_plane(self, point: Iterable) -> bool:
        if isinstance(point, np.ndarray) is False:
            point = np.array(point)

        return bool(point.dot(self.normal) + self.d == 0)

    def get_lcsys(self):
        if self.yvec is None:
            if sum(abs(self.normal) - np.array([0, 0, 1])) < 1e-5:
                vec1 = np.array([1, 0, 0])
            else:
                vec1 = np.array([0, 0, 1])

            self.yvec = unit_vector(calc_yvec(vec1, self.normal))

        xvec = unit_vector(calc_xvec(self.yvec, self.normal))

        return [xvec, self.yvec, self.normal]

    def get_points_in_lcsys_plane(self, p_dist: float = 1, plane: Plane = Plane.XY):
        csys = self.get_lcsys()
        p0 = self.point_in_plane
        vec_map = {
            Plane.XY: (0, 1),
            Plane.XZ: (0, 2),
            Plane.YZ: (1, 2),
        }
        i, j = vec_map.get(plane)
        vec2 = csys[i]
        vec3 = csys[j]

        p1 = p0 + vec2 * p_dist + vec3 * p_dist
        p2 = p0 - vec2 * p_dist + vec3 * p_dist
        p3 = p0 - vec2 * p_dist - vec3 * p_dist
        p4 = p0 + vec2 * p_dist - vec3 * p_dist
        return [p1, p2, p3, p4]

    def project_point_onto_plane(self, point: Iterable) -> np.ndarray:
        p = np.array(point)
        dist = p.dot(self.normal) + self.d
        return p - dist * self.normal


def linear_2dtransform_rotate(origin, point, degrees) -> np.ndarray:
    """
    Rotate a 2d point given an origin and a degree.

    :param origin: (x, y) coordinate of point of rotation
    :param point: (x, y) coordinate of point to rotate
    :param degrees: Degree rotation
    :return: Updated x, y coordinates based on rotation
    """
    theta = np.deg2rad(degrees)
    A = np.column_stack([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    v = np.array(point) - np.array(origin)
    #    return A @ v.T
    return np.array(origin) + A.dot(v.T)


def rot_matrix_alt(a, b, g, inverse=False, matr_tol=8):
    """

    :param a:
    :param b:
    :param g:
    :param inverse:
    :param matr_tol: Matrix decimal precision tolerance
    :return:
    """

    def R_x(al):
        return np.array([[1, 0, 0], [0, np.cos(al), -np.sin(al)], [0, np.sin(al), np.cos(al)]])

    def R_y(be):
        return np.array([[np.cos(be), 0, np.sin(be)], [0, 1, 0], [-np.sin(be), 0, np.cos(be)]])

    def R_z(ga):
        return np.array([[np.cos(ga), -np.sin(ga), 0], [np.sin(ga), np.cos(ga), 0], [0, 0, 1]])

    rot_mat = np.dot(R_z(g), np.dot(R_y(b), R_x(a)))
    rot_mat = np.around(rot_mat, matr_tol)
    if inverse:
        return rot_mat.T
    else:
        return rot_mat


def rotation_matrix_csys_rotate(csys1_in, csys2_in, inverse=False, use_quaternion=True):
    """
    Create a rotation matrix by providing two sets of coordinate systems defined by 2 vectors each (X- and Y-vectors)

    Resources:

        https://en.wikipedia.org/wiki/Rotation_matrix
        https://kieranwynn.github.io/pyquaternion/

    :param csys1_in: Coordinate system 1 defined by 2 vectors [LocalX, LocalY]
    :param csys2_in: Coordinate system 1 defined by 2 vectors [LocalX, LocalY]
    :param inverse: True if you want to reverse the applied rotations from csys1 to csys2
    :param use_quaternion: Create transformation matrix using the pyquaternion package
    :return: A rotation matrix ready to be employed for further use
    """

    # Ensure Consistent right-hand-rule
    csys1 = np.array([csys1_in[0], csys1_in[1], np.cross(csys1_in[0], csys1_in[1])]).astype(float)
    csys2 = np.array([csys2_in[0], csys2_in[1], np.cross(csys2_in[0], csys2_in[1])]).astype(float)

    if not np.allclose(np.dot(csys1, csys1.conj().transpose()), np.eye(3), rtol=1e-05, atol=1e-08):
        use_quaternion = False

    # Using PyQuaternion
    if use_quaternion:
        from pyquaternion import Quaternion

        q1 = Quaternion(matrix=csys1)
        q2 = Quaternion(matrix=csys2)
        q3 = q1 * q2

        if inverse:
            q3 = q3.inverse

        rm = q3.rotation_matrix
        return rm
    else:
        a = angle_between(csys1[0], csys2[0])
        b = angle_between(csys1[1], csys2[1])
        z = angle_between(csys1[2], csys2[2])

        rm_alt = rot_matrix_alt(a, b, z, inverse)
        return rm_alt


def angle_between(v1, v2):
    """
    Returns the angle in radians between vectors 'v1' and 'v2'

    source:

        https://stackoverflow.com/questions/2827393/angles-between-two-n-dimensional-vectors-in-python/13849249#13849249

    :param v1:
    :param v2:

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


def rot_matrix(normal, original_normal=(0, 0, 1)):
    """
    Creates a rotation matrix between two normals

    :param normal:
    :param original_normal:
    :return:
    """
    # Check if the two normals are similar
    if [True if abs(x - y) == 0 else False for x, y in zip(normal, original_normal)] == [True, True, True]:
        return np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    M = np.array(normal)
    N = np.array(original_normal)
    costheta = np.dot(M, N) / (np.linalg.norm(M) * np.linalg.norm(N))

    axis = np.cross(M, N) / np.linalg.norm(np.cross(M, N))
    if np.isnan(axis[0]) or np.isnan(axis[1]) or np.isnan(axis[2]):
        raise ValueError("Axis contains isnan members")
    c = costheta
    s = np.sqrt(1 - c * c)
    C = 1 - c
    x, y, z = axis[0], axis[1], axis[2]

    return np.array(
        [
            [x * x * C + c, x * y * C - z * s, x * z * C + y * s],
            [y * x * C + z * s, y * y * C + c, y * z * C - x * s],
            [z * x * C - y * s, z * y * C + x * s, z * z * C + c],
        ]
    )


def rotate_plane(points, normal, global_normal=(0, 0, 1)):
    """
    Method takes points from local plane rotates plane into global plane and returns points

    .. Note: Assumes rotation about origin (0,0,0)

    :param points:
    :param normal:
    :param global_normal:
    :return:
    """

    rmat = rot_matrix(normal, global_normal)
    newpoints = []
    for pid in points:
        point = points[pid].no.p
        newpoint = np.dot(rmat, point)
        newpoints.append(newpoint)
    return newpoints


def vector_length(vector):
    """
    This method takes in a np.array vector and returns the length
    of the vector.

    :param vector: A numpy array of a 3d vector
    :type vector: np.array
    :rtype: float
    """
    if len(vector) == 2:
        raise ValueError("Vector is not a 3d vector, but a 2d vector. Please consider vector_length_2d() instead")
    elif len(vector) != 3:
        raise ValueError(f"Vector is not a 3d vector. Vector array length: {len(vector)}")

    return float(np.linalg.norm(vector))


def vector_length_2d(vector):
    """
    This method takes in a np.array vector and returns the length
    of the vector.

    :param vector: A numpy array of a 2d vector
    :type vector: np.array
    :rtype: float
    """
    if len(vector) == 3:
        raise ValueError("Vector is not a 2d vector, but a 3d vector. Please consider vector_length() instead")
    elif len(vector) != 2:
        raise ValueError(f"Vector is not a 2d vector. Vector array length: {len(vector)}")

    return float(np.linalg.norm(vector))


def distfunc(x, point, A, B):
    """
    A function of x for the distance between point A on vector AB and arbitrary point C. X is a scalar multiplied with
    AB vector based on the distance to C.

    f(x) = np.sqrt(vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2)

    vector = C-(A+x*AB)

    :param x: Variable x
    :type x:
    :param point: Arbitrary point parallell to AB vector
    :type point:
    :param A: Point A on AB vector
    :type A:
    :param B: Point B on AB vector
    :type B:
    :return: Vector length
    :rtype:
    """
    AB = B - A
    return vector_length(point - (A + x * AB))


def sort_points_by_dist(p, points):
    return sorted(points, key=lambda x: vector_length(x - p))


def is_in_interval(value: float, interval_start: float, interval_end: float, incl_interval_ends: bool = False) -> bool:
    if incl_interval_ends:
        return interval_start <= value <= interval_end
    else:
        return interval_start < value < interval_end


def is_between_endpoints(p: np.ndarray, start: np.ndarray, end: np.ndarray, incl_endpoints: bool = False) -> bool:
    """Returns if point p is on the line between the points start and end"""
    if is_null_vector(p, start) or is_null_vector(p, end):
        if incl_endpoints:
            return True
        return False

    ab = end - start
    ap = p - start

    vec_fraction = get_vec_fraction(ap, ab)
    on_line_segment = is_in_interval(vec_fraction, 0.0, 1.0, incl_interval_ends=incl_endpoints)
    return is_parallel(ab, ap) and on_line_segment


def get_vec_fraction(vec: np.ndarray, reference_vec: np.ndarray) -> float:
    """Returns the fraction of the projection of vec onto reference_vec."""
    return np.dot(vec, reference_vec) / np.dot(reference_vec, reference_vec)


def point_on_line(start: np.ndarray, end: np.ndarray, point: np.ndarray) -> np.ndarray:
    """

    :param start: Start of line
    :param end: End of line
    :param point: Point
    :return:
    """
    ap = point - start
    ab = end - start
    result = start + get_vec_fraction(ap, ab) * ab
    return result


def is_null_vector(ab: np.array, cd: np.array, decimals=Settings.precision) -> bool:
    """Check if difference in vectors AB and CD is null vector"""
    return np.array_equal((cd - ab).round(decimals), np.zeros_like(ab))


def is_parallel(ab: np.array, cd: np.array, tol=Settings.point_tol) -> bool:
    """Check if vectors AB and CD are parallel"""
    return float(np.abs(np.sin(angle_between(ab, cd)))) < tol


def is_perpendicular(ab: np.array, cd: np.array, tol=Settings.point_tol) -> bool:
    """Returns if the vectors are perpendicular"""
    return float(np.abs(np.dot(ab, cd))) < tol


def is_angled(vector_1: np.ndarray, vector_2: np.ndarray) -> bool:
    """Returns true if 2 vectors is not perpendicular nor parallel to each other"""
    return not (is_perpendicular(vector_1, vector_2) or is_parallel(vector_1, vector_2))


def intersect_calc(a: np.ndarray, c: np.ndarray, ab: np.ndarray, cd: np.ndarray):
    """Function for evaluating an intersection point between two vector-lines (AB & CD).  The function returns
    variables s & t denoting the scalar value multiplied with the two vector equations A + s*AB = C + t*CD."""
    # Setting up the equation for use in linalg.lstsq
    matrix = np.array((ab, -cd)).T
    vec = c - a

    st = np.linalg.lstsq(matrix, vec, rcond=None)

    s = st[0][0]
    t = st[0][1]
    return s, t


def intersection_point(v1, v2):
    """
    Get the coordinate of the intersecting point
    :param v1:
    :param v2:
    :return:
    """
    is2d = len(list(v1[0])) == 2

    v1 = [np.array(list(v) + [0.0]) for v in list(v1)] if is2d else v1
    v2 = [np.array(list(v) + [0.0]) for v in list(v2)] if is2d else v2
    v1_ = v1[1] - v1[0]
    v2_ = v2[1] - v2[0]
    p1 = v1[0]
    p3 = v2[0]
    s, t = intersect_calc(p1, p3, v1_, v2_)
    res = p1 + s * v1_
    if is2d:
        return res[0], res[1]
    else:
        return res


def normalize(curve):
    if type(curve) is tuple:
        return (curve[0], [y / max(abs(curve[1])) for y in curve[1]])
    else:
        return [x / max(abs(curve)) for x in curve]


def is_point_inside_bbox(p, bbox, tol=1e-3) -> bool:
    """

    :param p: Point
    :param bbox: Bounding box
    :param tol: Tolerance
    :return:
    """
    if (
        bbox[0][0][0] - tol < p[0] < bbox[0][1][0] + tol
        and bbox[1][0][1] - tol < p[1] < bbox[1][1][1] + tol
        and bbox[2][0][2] - tol < p[2] < bbox[2][1][2] + tol
    ):
        return True
    else:
        return False


def points_in_cylinder(pt1, pt2, r, q):
    """

    :param pt1: Start point of cylinder
    :param pt2: End point of cylinder
    :param r: Radius of cylinder
    :param q:
    :return:
    """
    vec = pt2 - pt1
    const = r * np.linalg.norm(vec)
    if np.dot(q - pt1, vec) >= 0 >= np.dot(q - pt2, vec) and np.linalg.norm(np.cross(q - pt1, vec)) <= const:
        return True
    else:
        return False


def split(u, v, points):
    """

    :param u: Vector 1
    :param v: Vector 2
    :param points:
    :return:
    """

    # return points on left side of UV
    return [p for p in points if np.cross(p - u, v - u) < 0]


def extend(u, v, points):
    if not points:
        return []

    # find furthest point W, and split search to WV, UW
    w = min(points, key=lambda p: np.cross(p - u, v - u))
    p1, p2 = split(w, v, points), split(u, w, points)
    return extend(w, v, p1) + [w] + extend(u, w, p2)


def convex_hull(points):
    # find two hull points, U, V, and split to left and right search
    u = min(points, key=lambda p: p[0])
    v = max(points, key=lambda p: p[0])
    left, right = split(u, v, points), split(v, u, points)

    # find convex hull on each side
    return [v] + extend(u, v, left) + [u] + extend(v, u, right) + [v]


def is_coplanar(x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4) -> bool:
    """
    Python program to check if 4 points in a 3-D plane are Coplanar
    Function to find equation of plane.

    :param x1:
    :param y1:
    :param z1:
    :param x2:
    :param y2:
    :param z2:
    :param x3:
    :param y3:
    :param z3:
    :param x4:
    :param y4:
    :param z4:
    :return:
    """
    a1 = x2 - x1
    b1 = y2 - y1
    c1 = z2 - z1
    a2 = x3 - x1
    b2 = y3 - y1
    c2 = z3 - z1
    a = b1 * c2 - b2 * c1
    b = a2 * c1 - a1 * c2
    c = a1 * b2 - b1 * a2
    d = -a * x1 - b * y1 - c * z1
    # equation of plane is: a*x + b*y + c*z = 0 #
    # checking if the 4th point satisfies
    # the above equation
    if a * x4 + b * y4 + c * z4 + d == 0:
        # if(a * x + b * y + c * z + d < 0.0001 and a * x + b * y + c * z + d > -0.0001):
        # print("Coplanar")
        return True
    else:
        # print(a * x + b * y + c * z + d)
        return False
        # print("Not Coplanar")


def poly_area(x, y):
    """
    Shoelace formula based on stackoverflow

    https://stackoverflow.com/questions/24467972/calculate-area-of-polygon-given-x-y-coordinates

    :param x:
    :param y:
    :return:
    """
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def poly_area_from_list(coords):
    return poly_area(*zip(*coords))


def poly2d_center_of_gravity(polygon):
    """Calculates the center of gravity of a polygon described by a numpy array containing x,y,z coordinates."""
    sum_area = 0
    sum_cx = 0
    sum_cy = 0

    for i in range(len(polygon)):
        j = (i + 1) % len(polygon)
        area = polygon[i][0] * polygon[j][1] - polygon[j][0] * polygon[i][1]
        sum_area += area
        sum_cx += (polygon[i][0] + polygon[j][0]) * area
        sum_cy += (polygon[i][1] + polygon[j][1]) * area

    if sum_area == 0:
        return None

    cx = sum_cx / (3 * sum_area)
    cy = sum_cy / (3 * sum_area)

    return np.array([cx, cy])


def global_2_local_nodes(csys, origin, nodes, use_quaternion=True):
    """

    :param csys: List of tuples containing; [LocalX, LocalY]
    :param origin: Point of origin. Node object
    :param nodes: list of nodes
    :return: List of local node coordinates
    """
    from ada import Node

    global_csys = [(1, 0, 0), (0, 1, 0)]
    rmat = rotation_matrix_csys_rotate(global_csys, csys, use_quaternion=use_quaternion)

    if type(origin) is Node:
        origin = origin.p
    elif type(origin) in (list, tuple):
        origin = np.array(origin)

    if type(nodes[0]) is Node:
        nodes = [no.p for no in nodes]

    res = [np.dot(rmat, p - origin) for p in nodes]
    return res


def local_2_global_points(points, origin, xdir, normal):
    """
    A method for converting a list of points in a 2d coordinate system to global 3d coordinates

    :param normal: Normal to 2d plane
    :param origin: Origin of local coordinate system
    :param points: List of points in 2d coordinate system
    :param xdir: Local X-direction
    :return:
    """
    from ada.concepts.points import Node
    from ada.core.constants import X, Y

    if type(origin) is Node:
        origin = origin.p

    if type(points[0]) is Node:
        points = [no.p for no in points]

    points = [
        np.array(n, dtype=np.float64) if len(n) == 3 else np.array(list(n) + [0], dtype=np.float64) for n in points
    ]
    yvec = calc_yvec(xdir, normal)

    return transform3d([xdir, yvec], [X, Y], origin, points)


def transform3d(csys_1, csys_2, origin, points) -> List[np.ndarray]:
    """Transform points between coordinate systems"""
    rmat = rotation_matrix_csys_rotate(csys_1, csys_2, inverse=True)

    return [np.array(origin, dtype=np.float64) + np.dot(rmat, n) for n in points]


def normal_to_points_in_plane(points_) -> np.ndarray:
    """Get normal to the plane created by a list of points"""
    if len(points_) <= 2:
        raise ValueError("Insufficient number of points")

    # remove duplicate points
    set_data = set(tuple(x) for x in points_)
    if len(set_data) != len(points_):
        points = [np.array(y) for y in set_data]
    else:
        points = points_

    # take 3 arbitrary points and create a normal
    p1 = points[0]
    p2 = points[1]
    p3 = points[2]

    # These two vectors are in the plane
    v1 = p3 - p1
    v2 = p2 - p1

    if is_parallel(v1, v2) is True:
        for i in range(3, len(points)):
            p3 = points[i]
            v1 = p3 - p1
            if is_parallel(v1, v2) is False:
                break

    # the cross product is a vector normal to the plane
    n = np.array([x if abs(x) != 0.0 else 0.0 for x in list(np.cross(v1, v2))])
    if n.any() == 0.0:
        raise ValueError("Error in calculating plate normal")

    return np.array([x if abs(x) != 0.0 else 0.0 for x in list(unit_vector(n))])


def unit_vector(vector: np.ndarray) -> np.ndarray:
    """Returns the unit vector of a given vector"""
    norm = vector / np.linalg.norm(vector)
    if np.isnan(norm).any():
        raise VectorNormalizeError(f'Error trying to normalize vector "{vector}"')

    return norm


def is_clockwise(points) -> bool:
    """Return true if order of 2d points are sorted in a clockwise order"""
    psum = 0
    for p1, p2 in zip(points[:-1], points[1:]):
        psum += (p2[0] - p1[0]) * (p2[1] + p1[1])
    psum += (points[-1][0] - points[0][0]) * (points[-1][1] + points[0][1])
    return not float(psum) < 0


def calc_xvec(y_vec, z_vec):
    return np.cross(y_vec, z_vec)


def calc_yvec(x_vec, z_vec=None) -> np.ndarray:
    if z_vec is None:
        calc_zvec(x_vec)

    return np.cross(z_vec, x_vec)


def calc_zvec(x_vec, y_vec=None) -> np.ndarray:
    """Calculate Z-vector (up) from an x-vector (along beam) only."""
    from ada.core.constants import Y, Z

    if y_vec is None:
        z_vec = np.array(Z)
        a = angle_between(x_vec, z_vec)
        if a == np.pi or a == 0:
            z_vec = np.array(Y)
        return z_vec
    else:
        return np.cross(x_vec, y_vec)


def is_on_line(data):
    """Evaluate intersection point between two lines"""
    l, bm = data
    A, B = np.array(l[0]), np.array(l[1])
    AB = B - A
    C = bm.n1.p
    D = bm.n2.p
    CD = D - C

    if (vector_length(A - C) < 1e-5) is True and (vector_length(B - D) < 1e-5) is True:
        return None

    s, t = intersect_calc(A, C, AB, CD)
    AB_ = A + s * AB
    CD_ = C + t * CD
    if (vector_length(AB_ - CD_) < 1e-4) is True and s not in (0.0, 1.0):
        return list(AB_), bm
    else:
        return None


def projection_onto_line(point: np.ndarray, start: np.ndarray, end: np.ndarray) -> np.ndarray:
    """

    :param point: Point outside line
    :param start: Start node of line
    :param end: End node of line
    :return: Projection from n1 to p0 onto line. Returns projected line segment
    """

    v = end - start
    p = point - start
    angle = angle_between(v, p)
    t0 = np.linalg.norm(p) * np.cos(angle) * unit_vector(v)
    q = t0 - p
    return point + q



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem import StepImplicit

from ..write_loads import write_load
from .fields import create_field_output_str

if TYPE_CHECKING:
    from ada.concepts.spatial import Part


def step_static_lin_str(step: StepImplicit, part: Part) -> str:
    from ada.fem.exceptions.model_definition import (
        NoBoundaryConditionsApplied,
        NoLoadsApplied,
    )

    load_str = "\n".join(list(map(write_load, step.loads)))
    if len(step.loads) == 0:
        raise NoLoadsApplied(f"No loads are applied in step '{step}'")
    load = step.loads[0]
    all_boundary_conditions = part.fem.bcs
    assembly = part.get_assembly()
    if assembly != part:
        for bc in part.get_assembly().fem.bcs:
            if bc not in all_boundary_conditions:
                all_boundary_conditions.append(bc)

    if len(all_boundary_conditions) == 0:
        raise NoBoundaryConditionsApplied("No boundary condition is found for the specified model")

    bc_str = ""
    for bc in all_boundary_conditions:
        bc_str += f"_F(CHARGE={bc.name}),"

    sec_str = ""
    if len(part.fem.sections.lines) > 0 or len(part.fem.sections.shells) > 0:
        sec_str = "\n    CARA_ELEM=element,"

    field_str = create_field_output_str(step)

    return f"""
{load_str}

result = MECA_STATIQUE(
    MODELE=model,
    CHAM_MATER=material,{sec_str}
    EXCIT=({bc_str}_F(CHARGE={load.name}))
)

{field_str}

IMPR_RESU(
    RESU=_F(RESULTAT=result),
    UNITE=80
)

"""



from __future__ import annotations

import json
import os
import pathlib
import subprocess
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List

import meshio
import numpy as np

from ada.config import get_logger
from ada.fem.formats.general import FEATypes
from ada.visualize.femviz import get_edges_and_faces_from_meshio, magnitude

from ..formats.abaqus.results import read_abaqus_results
from ..formats.calculix.results import read_calculix_results
from ..formats.code_aster.results import read_code_aster_results
from ..formats.sesam.results import read_sesam_results
from .eigenvalue import EigenDataSummary

if TYPE_CHECKING:
    from ada import Assembly
    from ada.visualize.concept import PartMesh, VisMesh

logger = get_logger()


class Results:
    res_map = {
        ".rmed": (read_code_aster_results, FEATypes.CODE_ASTER),
        ".frd": (read_calculix_results, FEATypes.CALCULIX),
        ".odb": (read_abaqus_results, FEATypes.ABAQUS),
        ".sin": (read_sesam_results, FEATypes.SESAM),
    }

    def __init__(
        self,
        res_path,
        name: str = None,
        fem_format: str | FEATypes = None,
        assembly: None | Assembly = None,
        palette=None,
        output=None,
        overwrite=True,
        metadata=None,
        import_mesh=False,
    ):
        if isinstance(fem_format, str):
            fem_format = FEATypes.from_str(fem_format)

        self._name = name
        self._visualizer = ResultsMesh(palette, fem_format=fem_format, parent=self)
        self._eigen_mode_data = None
        self._fem_format = fem_format
        self._assembly = assembly
        self._output = output
        self._overwrite = overwrite
        self._metadata = metadata if metadata is not None else dict()
        self._results_file_path = pathlib.Path(res_path) if res_path is not None else None
        self._user_data = dict()
        self._history_output = None
        self._import_mesh = import_mesh
        if res_path is not None:
            self.load_data_from_result_file(self.results_file_path)
            if self.results_file_path.exists():
                self._last_modified = os.path.getmtime(str(self.results_file_path))
            else:
                self._last_modified = None

    def load_data_from_result_file(self, file_ref=None, overwrite=False):
        file_ref = self.results_file_path if file_ref is None else file_ref

        if file_ref is None:
            return None

        if file_ref.exists() is False:
            return None

        mesh = self._get_results_from_result_file(file_ref)

        if mesh is None:
            return None

        if self._import_mesh is False:
            return None
        print(f'Importing meshio.Mesh from result file "{file_ref}"')
        self.result_mesh.add_results(mesh)

    def _get_results_from_result_file(self, file_ref, overwrite=False):
        file_ref = pathlib.Path(file_ref)
        suffix = file_ref.suffix.lower()

        res_reader, fem_format = Results.res_map.get(suffix, (None, None))

        if res_reader is None:
            logger.error(f'Results class currently does not support filetype "{suffix}"')
            return None

        self.fem_format = fem_format

        return res_reader(self, file_ref, overwrite)

    def save_output(self, dest_file) -> None:
        if self.output is None or self.output.stdout is None:
            print("No output is found")
            return None
        dest_file = pathlib.Path(dest_file)

        os.makedirs(dest_file.parent, exist_ok=True)
        with open(dest_file, "w") as f:
            f.write(self.output.stdout)

    def save_results_to_json(self, dest_file):
        dest_file = pathlib.Path(dest_file).with_suffix(".json")
        res = dict(
            name=self.name,
            fem_format=self.fem_format,
            eigen_mode_data=self.eigen_mode_data.to_dict(),
            metadata=self.metadata,
            last_modified=self.last_modified,
        )
        with open(dest_file, "w") as f:
            try:
                json.dump(res, f, indent=4)
            except TypeError as e:
                raise TypeError(e)

    def save_results_to_excel(self, dest_file, filter_components_by_name=None):
        """This method is just a sample for how certain results can easily be exported to Excel"""

        try:
            import xlsxwriter
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "xlsxwriter must be installed to export to xlsx.\n"
                'To install you can use "conda install -c conda-forge xlsxwriter"'
            )

        dest_file = pathlib.Path(dest_file).with_suffix(".xlsx")

        workbook = xlsxwriter.Workbook(dest_file)
        worksheet = workbook.add_worksheet()

        worksheet.write("A1", "Step")
        worksheet.write("B1", "Element")
        worksheet.write("C1", "ForceComponent")
        worksheet.write("D1", "Value")
        i = 2
        for step in self.history_output.steps:
            for el_name, el in step.element_data.items():
                el: ElementDataOutput
                for force_name, force in el.forces.items():
                    if filter_components_by_name is not None:
                        skip_it = False
                        if force.name.lower() not in [x.lower() for x in filter_components_by_name]:
                            skip_it = True
                        if skip_it:
                            continue
                    worksheet.write(f"A{i}", step.name)
                    worksheet.write(f"B{i}", el_name)
                    worksheet.write(f"C{i}", force.name)
                    worksheet.write(f"D{i}", force.final_force)
                    i += 1

        workbook.close()

    def to_vis_mesh(self, data_type: str = None, name: str = "AdaFEM") -> VisMesh | None:
        from ada.visualize.concept import VisMesh

        name = self.assembly.name if self.assembly is not None else name
        pm = self.result_mesh.to_part_mesh(name=name, data_type=data_type)
        if len(pm.id_map) == 0:
            logger.warning("Created Part mesh contains no object meshes")
            return None
        project = self.assembly.metadata.get("project", "DummyProject") if self.assembly is not None else "DummyProject"
        return VisMesh(name=name, project=project, world=[pm], meta=None)

    @property
    def name(self):
        return self._name

    @property
    def fem_format(self) -> FEATypes:
        return self._fem_format

    @fem_format.setter
    def fem_format(self, value: FEATypes):
        if isinstance(value, str):
            value = FEATypes.from_str(value)
            if value is None:
                raise ValueError(f'Unsupported FEA Type "{value}"')
        self._fem_format = value

    @property
    def last_modified(self):
        return self._last_modified

    @last_modified.setter
    def last_modified(self, value):
        self._last_modified = value

    @property
    def output(self) -> subprocess.CompletedProcess:
        return self._output

    @property
    def results_file_path(self):
        return self._results_file_path

    @results_file_path.setter
    def results_file_path(self, value):
        self._results_file_path = value

    @property
    def result_mesh(self) -> ResultsMesh:
        return self._visualizer

    @property
    def assembly(self) -> "Assembly":
        return self._assembly

    @property
    def eigen_mode_data(self) -> EigenDataSummary:
        return self._eigen_mode_data

    @eigen_mode_data.setter
    def eigen_mode_data(self, value: EigenDataSummary):
        self._eigen_mode_data = value

    @property
    def history_output(self) -> ResultsHistoryOutput:
        return self._history_output

    @history_output.setter
    def history_output(self, value: ResultsHistoryOutput):
        self._history_output = value

    @property
    def metadata(self):
        return self._metadata

    @property
    def user_data(self) -> dict:
        return self._user_data

    def _repr_html_(self):
        from IPython.display import display
        from ipywidgets import HBox, VBox

        if self.result_mesh.renderer is None:
            res = self.result_mesh.build_renderer()
        else:
            res = True

        if res is False:
            print("No ")
            return

        p3s_renderer = self.result_mesh.renderer
        display(HBox([VBox([HBox(p3s_renderer.controls), p3s_renderer.renderer]), p3s_renderer.html]))

    def __repr__(self):
        return f"Results({self._fem_format}, {self._results_file_path.name})"


@dataclass
class ElemForceComp:
    name: str
    data: List[tuple]

    @property
    def final_force(self):
        return self.data[-1][-1]


@dataclass
class ElementDataOutput:
    name: str
    displacements: dict[int, List[tuple]] = field(default_factory=dict)
    forces: dict[int, ElemForceComp] = field(default_factory=dict)

    @property
    def final_displ(self):
        return {x: y[-1][-1] for x, y in self.displacements.items()}

    @property
    def final_forces(self):
        return {x: y.data[-1][-1] for x, y in self.forces.items()}


@dataclass
class FEMDataOutput:
    name: str
    data: List[tuple]


@dataclass
class HistoryStepDataOutput:
    name: str
    step_type: str
    element_data: Dict[str, ElementDataOutput] = field(default_factory=dict)
    fem_data: Dict[str, FEMDataOutput] = field(default_factory=dict)


@dataclass
class ResultsHistoryOutput:
    steps: List[HistoryStepDataOutput] = field(default_factory=list)


@dataclass
class ResultsMesh:
    palette: list[tuple]
    parent: Results
    fem_format: str
    renderer: object = None
    render_sets: object = None
    mesh: meshio.Mesh = None
    undeformed_mesh: None | object = None
    deformed_mesh: tuple[object, object, object] = None
    point_data: list = field(default_factory=list)
    cell_data: list = field(default_factory=list)

    vertices: np.ndarray = None
    edges: np.ndarray = None
    faces: np.ndarray = None

    def __post_init__(self):
        self.palette = [(0, 149 / 255, 239 / 255), (1, 0, 0)] if self.palette is None else self.palette

    def add_results(self, mesh: meshio.Mesh):
        self.mesh = mesh
        self.vertices = np.asarray(mesh.points, dtype="float32")

        edges, faces = get_edges_and_faces_from_meshio(mesh)
        self.edges = np.asarray(edges, dtype="uint16").ravel()
        self.faces = np.asarray(faces, dtype="uint16").ravel()

        for n in mesh.point_data.keys():
            self.point_data.append(n)

        for n in mesh.cell_data.keys():
            self.cell_data.append(n)

    def build_renderer(self) -> bool:
        from ipywidgets import Dropdown

        from ada.visualize.renderer_pythreejs import MyRenderer

        self.renderer = MyRenderer()
        if len(self.point_data) == 0:
            return False
        if self.fem_format == FEATypes.CODE_ASTER:
            data = [x for x in self.point_data if "DISP" in x][-1]
        elif self.fem_format == FEATypes.CALCULIX:
            data = [x for x in self.point_data if "U" in x][-1]
        else:
            raise NotImplementedError(f'Support for analysis_type "{self.fem_format}"')

        self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
        i = self.point_data.index(data)
        self.render_sets = Dropdown(
            options=self.point_data, value=self.point_data[i], tooltip="Select a set", disabled=False
        )
        self.render_sets.observe(self.on_changed_point_data_set, "value")
        self.renderer.controls.pop()
        self.renderer.controls.append(self.render_sets)
        return True

    def colorize_data(self, data, func=magnitude):
        res = [func(d) for d in data]
        sorte = sorted(res)
        min_r = sorte[0]
        max_r = sorte[-1]

        start = np.array(self.palette[0])
        end = np.array(self.palette[-1])

        def curr_p(t):
            return start + (end - start) * t / (max_r - min_r)

        colors = np.asarray([curr_p(x) for x in res], dtype="float32")
        return colors

    def create_viz_geom(self, data_type, displ_data=False, renderer: object = None) -> None:
        from ada.visualize.renderer_pythreejs import MyRenderer
        from ada.visualize.threejs_utils import (
            edges_to_mesh,
            faces_to_mesh,
            vertices_to_mesh,
        )

        default_vertex_color = (8, 8, 8)

        data = np.asarray(self.mesh.point_data[data_type], dtype="float32")
        colors = self.colorize_data(data)

        if renderer is None:
            renderer = MyRenderer()
            self.renderer = renderer

        # deformations
        if displ_data is True:
            vertices = np.asarray([x + u[:3] for x, u in zip(self.vertices, data)], dtype="float32")
            if self.undeformed_mesh is None:
                dark_grey = (0.66, 0.66, 0.66)
                white_color = np.asarray([dark_grey for x in self.vertices], dtype="float32")
                o_mesh = faces_to_mesh("undeformed", self.vertices, self.faces, white_color, opacity=0.5)
                self.undeformed_mesh = o_mesh
                renderer._displayed_non_pickable_objects.add(o_mesh)
        else:
            vertices = self.vertices
            if self.undeformed_mesh is not None:
                renderer._displayed_non_pickable_objects.remove(self.undeformed_mesh)
                self.undeformed_mesh = None

        vertices = np.array(vertices, dtype=np.float32)

        # Colours
        mesh = faces_to_mesh("deformed", vertices, self.faces, colors)
        points = vertices_to_mesh("deformed_vertices", vertices, default_vertex_color)
        lines = edges_to_mesh("deformed_lines", vertices, self.edges, default_vertex_color)

        if self.deformed_mesh is None:
            self.deformed_mesh = (mesh, points, lines)
            renderer.displayed_pickable_objects.add(mesh)
            renderer.displayed_pickable_objects.add(points)
            renderer.displayed_pickable_objects.add(lines)
            renderer.build_display(camera_type="perspective")
        else:
            face_geom = self.deformed_mesh[0].geometry
            face_geom.attributes["position"].array = vertices
            face_geom.attributes["index"].array = self.faces
            face_geom.attributes["color"].array = colors

            point_geom = self.deformed_mesh[1].geometry
            point_geom.attributes["position"].array = vertices

            edge_geom = self.deformed_mesh[2].geometry
            edge_geom.attributes["position"].array = vertices
            edge_geom.attributes["index"].array = self.edges

    def on_changed_point_data_set(self, p):
        data = p["new"]
        if self.fem_format == FEATypes.CODE_ASTER:
            is_displ = True if "DISP" in data else False
            if "point_tags" in data:
                print("\r" + "Point Tags are not a valid display value" + 10 * " ", end="")
                return None
        elif self.fem_format == FEATypes.CALCULIX:
            is_displ = True if "U" in data else False
        else:
            return None

        if is_displ:
            self.create_viz_geom(data, displ_data=True, renderer=self.renderer)
        else:
            self.create_viz_geom(data, renderer=self.renderer)

    def to_part_mesh(self, name: str, data_type: str = None) -> PartMesh:
        from ada.visualize.concept import ObjectMesh, PartMesh

        if data_type is not None:
            step_names = [data_type]
        else:
            step_names = self.point_data

        id_map = dict()
        for step in step_names:
            data = np.asarray(self.mesh.point_data[step], dtype="float32")
            if step.lower() in ["forc", "stress"]:
                vertices = np.asarray([x for x, u in zip(self.vertices, data)], dtype="float32")
            else:
                try:
                    vertices = np.asarray([x + u[:3] for x, u in zip(self.vertices, data)], dtype="float32")
                except IndexError:
                    logger.warning(f"Data step {step} contains invalid data. Skipping")
                    continue
            try:
                colors = self.colorize_data(data)
            except IndexError:
                logger.warning(f"Data step {step} was unable to colorize data. Skipping")
                continue

            faces = self.faces
            edges = self.edges

            id_map[step] = ObjectMesh(
                guid=step,
                faces=faces.astype(int),
                position=vertices.flatten().astype(float),
                normal=None,
                color=None,
                vertex_color=colors.flatten().astype(float),
                edges=edges,
                instances=None,
                # id_sequence=dict()
            )

        return PartMesh(name=name, id_map=id_map)


def get_fem_stats(fem_file, dest_md_file, data_file="data.json"):
    """

    :param fem_file:
    :param dest_md_file:
    :param data_file: Destination of data.json file (keeping track of last modified status etc..)
    """
    import json
    import os

    from ada import Assembly
    from ada.fem.utils import get_eldata

    dest_md_file = pathlib.Path(dest_md_file)
    data_file = pathlib.Path(data_file)
    a = Assembly()
    a.read_fem(fem_file)

    out_str = ""

    for name, part in a.parts.items():
        fem = part.fem
        r = get_eldata(fem_source=fem)
        if len(r.keys()) == 0:
            continue
        out_str += f"* **{name}**: ("

        el_data = ""
        for el_type, el_num in r.items():
            el_data += f"{el_type}: {el_num}, "

        out_str += el_data[:-2] + ")\n"

    os.makedirs(dest_md_file.parent, exist_ok=True)

    with open(dest_md_file, "w") as f:
        f.write(out_str)

    if data_file.exists():
        with open(data_file, "r") as f:
            data = json.load(f)
    else:
        data = dict()
    print(data)



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.visualize.concept import PartMesh, VisMesh
from ada.visualize.config import ExportConfig

from .write_objects_to_mesh import obj_to_mesh
from .write_part_to_mesh import generate_meta

if TYPE_CHECKING:
    from ada.concepts.connections import JointBase


def export_joint_to_assembly_mesh(joint: "JointBase", export_config: ExportConfig) -> VisMesh:
    all_obj = [obj for obj in joint.beams]
    all_obj_num = len(all_obj)

    print(f"Exporting {all_obj_num} physical objects to custom json format.")
    obj_num = 1

    id_map = dict()
    for obj in all_obj:
        res = obj_to_mesh(obj, export_config)
        if res is None:
            continue
        id_map[obj.guid] = res
        print(f'Exporting "{obj.name}" ({obj_num} of {all_obj_num})')

    meta = generate_meta(joint.parent, export_config) if joint.parent is not None else None

    return VisMesh(
        name=joint.name,
        project=joint.metadata.get("project", "DummyProject"),
        world=[PartMesh(joint.name, id_map=id_map)],
        meta=meta,
    )



from __future__ import annotations

import traceback
from dataclasses import dataclass
from itertools import chain
from typing import Iterable, List

import numpy as np

from ada import Assembly, Beam, Node, Part, Pipe, PipeSegStraight, Plate, PrimCyl
from ada.config import get_logger

from .utils import Counter
from .vector_utils import EquationOfPlane, intersect_calc, is_parallel, vector_length

logger = get_logger()


def basic_intersect(bm: Beam, margins, all_parts: [Part]):
    if bm.section.type == "gensec":
        return bm, []
    try:
        vol = bm.bbox().minmax
    except ValueError as e:
        logger.error(f"Intersect bbox skipped: {e}\n{traceback.format_exc()}")
        return None
    vol_in = [x for x in zip(vol[0], vol[1])]
    beams = filter(
        lambda x: x != bm,
        chain.from_iterable([p.beams.get_beams_within_volume(vol_in, margins=margins) for p in all_parts]),
    )
    return bm, beams


def beam_cross_check(bm1: Beam, bm2: Beam, outofplane_tol=0.1):
    """Calculate intersection of beams and return point, s, t"""
    p_check = is_parallel
    i_check = intersect_calc
    v_len = vector_length
    a = bm1.n1.p
    b = bm1.n2.p
    c = bm2.n1.p
    d = bm2.n2.p

    ab = b - a
    cd = d - c

    s, t = i_check(a, c, ab, cd)

    ab_ = a + s * ab
    cd_ = c + t * cd

    if p_check(ab, cd):
        logger.debug(f"beams {bm1} {bm2} are parallel")
        return None

    if v_len(ab_ - cd_) > outofplane_tol:
        logger.debug("The two lines do not intersect within given tolerances")
        return None

    return ab_, s, t


def are_beams_connected(bm1: Beam, beams: List[Beam], out_of_plane_tol, point_tol, nodes, nmap) -> None:
    # TODO: Function should be renamed, or return boolean. Unclear what the function does at the moment
    for bm2 in beams:
        if bm1 == bm2:
            continue
        res = beam_cross_check(bm1, bm2, out_of_plane_tol)
        if res is None:
            continue
        point, s, t = res
        t_len = (abs(t) - 1) * bm2.length
        s_len = (abs(s) - 1) * bm1.length
        if t_len > bm2.length / 2 or s_len > bm1.length / 2:
            continue
        if point is not None:
            new_node = Node(point)
            n = nodes.add(new_node, point_tol=point_tol)
            if n not in nmap.keys():
                nmap[n] = [bm1]
            if bm1 not in nmap[n]:
                nmap[n].append(bm1)
            if bm2 not in nmap[n]:
                nmap[n].append(bm2)


def are_plates_touching(pl1: Plate, pl2: Plate, tol=1e-3):
    """Check if two plates are within tolerance of each other"""
    from ..occ.utils import compute_minimal_distance_between_shapes

    dss = compute_minimal_distance_between_shapes(pl1.solid(), pl2.solid())
    if dss.Value() <= tol:
        return dss

    return None


def filter_away_beams_along_plate_edges(pl: Plate, beams: Iterable[Beam]) -> List[Beam]:
    corners = [n for n in pl.poly.points3d]

    # filter away all beams with both ends on any of corner points of the plate
    beams_not_along_plate_edge = []

    for bm in beams:
        t1 = tuple(bm.n1.p)
        t2 = tuple(bm.n2.p)

        if t1 in corners:
            cindex = corners.index(t1)

            nextp = corners[0] if cindex == len(corners) - 1 else corners[cindex + 1]
            prevp = corners[cindex - 1]

            if t2 == nextp or t2 == prevp:
                continue
        beams_not_along_plate_edge.append(bm)

    return beams_not_along_plate_edge


def filter_beams_along_plate_edges(pl: Plate, beams: Iterable[Beam]):
    from .vector_utils import is_clockwise, is_on_line

    corners = [n for n in pl.poly.points3d]
    corners += [corners[0]]
    if is_clockwise(corners):
        corners.reverse()

    # Evalute Corner Points
    crossing_beams = []
    for s, e in zip(corners[:-1], corners[1:]):
        li = (s, e)
        res = [x for x in map(is_on_line, [(li, bm) for bm in beams]) if x is not None]
        crossing_beams += filter(lambda x: x not in crossing_beams, [r[1] for r in res])

    return crossing_beams


def find_beams_connected_to_plate(pl: Plate, beams: List[Beam]) -> List[Beam]:
    """Return all beams with their midpoints inside a specified plate for a given list of beams"""
    from ada.concepts.containers import Nodes

    nid = Counter(1)
    nodes = Nodes([Node((bm.n2.p + bm.n1.p) / 2, next(nid), refs=[bm]) for bm in beams])
    res = nodes.get_by_volume(pl.bbox().p1, pl.bbox().p2)

    all_beams_within = list(chain.from_iterable([r.refs for r in res]))
    return all_beams_within


def penetration_check(part: Part):
    a = part.get_assembly()
    cog = part.nodes.vol_cog
    normal = part.placement.zdir
    for p in a.get_all_subparts():
        for pipe in p.pipes:
            for segment in pipe.segments:
                if type(segment) is PipeSegStraight:
                    assert isinstance(segment, PipeSegStraight)
                    p1, p2 = segment.p1, segment.p2
                    v1 = (p1.p - cog) * normal
                    v2 = (p2.p - cog) * normal
                    if np.dot(v1, v2) < 0:
                        part.add_penetration(
                            PrimCyl(f"{p.name}_{pipe.name}_{segment.name}_pen", p1.p, p2.p, pipe.section.r + 0.1)
                        )


@dataclass
class PipeClash:
    seg: PipeSegStraight
    plate: Plate

    @staticmethod
    def pipe_penetration_check(a: Assembly) -> list[PipeClash]:
        plates = list(a.get_all_physical_objects(by_type=Plate))
        pipes = list(a.get_all_physical_objects(by_type=Pipe))
        pipe_segments = []
        for pipe in pipes:
            pipe_segments += list(filter(lambda x: isinstance(x, PipeSegStraight), pipe.segments))

        clashes = []

        for seg in pipe_segments:
            p1 = seg.p1.p
            p2 = seg.p2.p
            for plate in plates:
                origin = plate.placement.origin
                normal = plate.placement.zdir

                v1 = (p1 - origin) * normal
                v2 = (p2 - origin) * normal
                is_clashing = np.dot(v1, v2) < 0
                if is_clashing:
                    print(f"{seg.name=} {is_clashing=} with {plate.name=}")
                    clashes.append(PipeClash(seg, plate))
        return clashes

    def reinforce_plate_pipe_pen(self, add_to_layer: str = None):
        seg = self.seg
        plate = self.plate

        p1 = seg.p1.p
        p2 = seg.p2.p

        pipe = seg.parent
        part = plate.parent

        # Cut away in plate and stringers here
        name = f"{plate.name}_{pipe.name}_{seg.name}_pen"
        part.add_penetration(PrimCyl(name, p1, p2, seg.section.r + 0.1), add_to_layer=add_to_layer)

        # specify reinforcement here
        reinforce_name = Counter(prefix=f"{plate.name}_{pipe.name}_{seg.name}_reinf_")

        eop = EquationOfPlane(plate.placement.origin, plate.placement.zdir, plate.placement.ydir)
        xdir, ydir, zdir = eop.get_lcsys()

        pp = eop.project_point_onto_plane(p1) + plate.t * plate.placement.zdir

        dist = 3 * seg.section.r

        bm_p1 = pp - dist * xdir - dist * ydir
        bm_p2 = pp + dist * xdir - dist * ydir
        bm_p3 = pp + dist * xdir + dist * ydir
        bm_p4 = pp - dist * xdir + dist * ydir

        part.add_beam(Beam(next(reinforce_name), bm_p1, bm_p2, "HP140x8"), add_to_layer=add_to_layer)
        part.add_beam(Beam(next(reinforce_name), bm_p2, bm_p3, "HP140x8"), add_to_layer=add_to_layer)
        part.add_beam(Beam(next(reinforce_name), bm_p3, bm_p4, "HP140x8"), add_to_layer=add_to_layer)
        part.add_beam(Beam(next(reinforce_name), bm_p4, bm_p1, "HP140x8"), add_to_layer=add_to_layer)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ada.ifc.utils import create_guid

if TYPE_CHECKING:
    from ada import Assembly, Beam, Part, Pipe, Plate, Shape, Wall


@dataclass
class Group:
    name: str
    members: list[Part | Beam | Plate | Wall | Pipe | Shape]
    parent: Part | Assembly
    description: str = ""
    guid: str = field(default_factory=create_guid)

    def to_part(self, name: str):
        p = Part(name)
        for mem in self.members:
            p.add_object(mem)
        return p



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.core.vector_utils import unit_vector, vector_length
from ada.ifc.utils import (
    create_axis,
    create_guid,
    create_ifc_placement,
    create_local_placement,
    ifc_p,
    to_real,
)

if TYPE_CHECKING:
    from ada import Assembly
    from ada.concepts.primitives import (
        BSplineSurfaceWithKnots,
        RationalBSplineSurfaceWithKnots,
    )
    from ada.ifc.store import IfcStore


def generate_extruded_area_solid_prod_def(ifc_store: IfcStore, p_start, p_end, section):
    f = ifc_store.f

    body_context = ifc_store.get_context("Body")
    axis_context = ifc_store.get_context("Axis")

    axis_representation = create_axis(f, [p_start, p_end], axis_context)

    section_profile = ifc_store.get_profile_def(section)
    if section_profile is None:
        raise ValueError("Section profile not found")

    body = create_extruded_body(f, p_start, p_end, section_profile)

    body_representation = f.createIfcShapeRepresentation(body_context, "Body", "SweptSolid", [body])
    return f.create_entity(
        "IfcProductDefinitionShape",
        Name=None,
        Description=None,
        Representations=[axis_representation, body_representation],
    )


def create_extruded_body(f, p1, p2, section_profile):
    xdir = to_real(unit_vector(p2.p - p1.p))
    ifcdir = f.createIfcDirection(xdir)
    extrusion_placement = create_ifc_placement(f, (0.0, 0.0, 0.0), (0.0, 0.0, 1.0), (1.0, 0.0, 0.0))
    seg_l = vector_length(p2.p - p1.p)

    return f.createIfcExtrudedAreaSolid(section_profile, extrusion_placement, ifcdir, seg_l)


def add_bsplinesurface_to_ifc(
    surface: BSplineSurfaceWithKnots | RationalBSplineSurfaceWithKnots, assembly: Assembly, parent_guid: str = None
):
    """A temporary function to add a B-Spline surface to an IFC file. To be integrated into ifcstore.sync()"""
    f = assembly.ifc_store.f
    context = assembly.ifc_store.get_context("Body")
    ifc_surface = surface.to_ifcopenshell(f)

    # IfcPolyLine
    p11 = ifc_p(f, (0, 0, 0))
    p21 = ifc_p(f, (0, 0, 1))
    poly_line = f.create_entity("IFCPOLYLINE", (p11, p21))

    # List of vertex points
    p1 = ifc_p(f, (0, 0, 0))
    p2 = ifc_p(f, (0, 0, 1))
    vp1 = f.create_entity("IfcVertexPoint", p1)
    vp2 = f.create_entity("IfcVertexPoint", p2)

    # List of edge curves
    edge_curve_1 = f.create_entity("IFCEDGECURVE", vp1, vp2, poly_line, True)

    # List of orient edges
    orient_edge_1 = f.create_entity("IFCORIENTEDEDGE", None, None, edge_curve_1, True)

    edge_loop = f.create_entity("IFCEDGELOOP", (orient_edge_1,))
    outer_bound = f.create_entity("IFCFACEOUTERBOUND", edge_loop, True)
    advanced_face = f.create_entity("IFCADVANCEDFACE", (outer_bound,), ifc_surface, True)
    closed_shell = f.create_entity("IFCCLOSEDSHELL", (advanced_face,))
    advanced_brep = f.create_entity("IFCADVANCEDBREP", closed_shell)
    shape_rep = f.create_entity("IFCSHAPEREPRESENTATION", context, "Body", "AdvancedBrep", (advanced_brep,))
    prod_def_shape = f.create_entity("IFCPRODUCTDEFINITIONSHAPE", None, None, (shape_rep,))
    local_place = create_local_placement(f)
    bldg_el_proxy = f.create_entity(
        "IFCBUILDINGELEMENTPROXY",
        create_guid(),
        assembly.ifc_store.owner_history,
        "BuildingElementProxy",
        None,
        None,
        local_place,
        prod_def_shape,
        None,
        "NOTDEFINED",
    )
    assembly.ifc_store.writer.add_related_elements_to_spatial_container([bldg_el_proxy], parent_guid)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ada import Beam
from ada.concepts.curves import CurveRevolve
from ada.config import get_logger
from ada.core.vector_utils import calc_yvec, vector_length

from .read_beam_section import import_section_from_ifc
from .read_materials import read_material
from .reader_utils import (
    get_associated_material,
    get_placement,
    get_point,
    get_swept_area,
)

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def import_ifc_beam(ifc_elem, name, ifc_store: IfcStore) -> Beam:
    from .exceptions import NoIfcAxesAttachedError

    mat_ref = get_associated_material(ifc_elem)

    mat_name = mat_ref.Material.Name if hasattr(mat_ref, "Material") else mat_ref.Name
    mat = ifc_store.assembly.get_by_name(mat_name)
    if mat is None:
        mat = read_material(mat_ref, ifc_store)

    swept_area = get_swept_area(ifc_elem)
    sec = import_section_from_ifc(swept_area, units=ifc_store.assembly.units)

    axes = [rep for rep in ifc_elem.Representation.Representations if rep.RepresentationIdentifier == "Axis"]

    if len(axes) != 1:
        raise NoIfcAxesAttachedError("Number of axis objects attached to IfcBeam is not 1")
    if len(axes[0].Items) != 1:
        raise ValueError("Number of items objects attached to axis is not 1")

    axis = axes[0].Items[0]
    if axis.is_a("IfcPolyline") and len(axis.Points) != 2:
        return import_polyline_beam(ifc_elem, axis, name, sec, mat, ifc_store)
    elif axis.is_a("IfcTrimmedCurve"):
        return import_revolved_beam(ifc_elem, axis, name, sec, mat, ifc_store)
    else:
        return import_straight_beam(ifc_elem, axis, name, sec, mat, ifc_store)


def get_beam_geom(ifc_elem, ifc_settings):
    # from .read_shapes import get_ifc_geometry
    # pdct_shape, colour, alpha = get_ifc_geometry(ifc_elem, ifc_settings)

    bodies = [rep for rep in ifc_elem.Representation.Representations if rep.RepresentationIdentifier == "Body"]
    if len(bodies) != 1:
        raise ValueError("Number of body objects attached to element is not 1")
    if len(bodies[0].Items) != 1:
        raise ValueError("Number of items objects attached to body is not 1")

    body = bodies[0].Items[0]
    if len(body.StyledByItem) > 0:
        style = body.StyledByItem[0].Styles[0].Styles[0].Styles[0]
        colour = (int(style.SurfaceColour.Red), int(style.SurfaceColour.Green), int(style.SurfaceColour.Blue))
        print(colour)


def import_straight_beam(ifc_elem, axis, name, sec, mat, ifc_store: IfcStore) -> Beam:
    p1_loc = axis.Points[0].Coordinates
    p2_loc = axis.Points[1].Coordinates

    ifc_axis_2_place3d = ifc_elem.ObjectPlacement.RelativePlacement
    origin = ifc_axis_2_place3d.Location.Coordinates

    local_z = np.array(ifc_axis_2_place3d.Axis.DirectionRatios)
    local_x = np.array(ifc_axis_2_place3d.RefDirection.DirectionRatios)
    local_y = calc_yvec(local_x, local_z)

    # res = transform3d([local_x, local_y, local_z], [X, Y], origin, [p1_loc, p2_loc])
    vlen = vector_length(np.array(p2_loc) - np.array(p1_loc))

    p1 = origin
    p2 = np.array(p1) + local_z * vlen

    return Beam(
        name,
        p1,
        p2,
        sec=sec,
        mat=mat,
        up=local_y,
        guid=ifc_elem.GlobalId,
        ifc_store=ifc_store,
        units=ifc_store.assembly.units,
    )


def import_revolved_beam(ifc_elem, axis, name, sec, mat, ifc_store: IfcStore) -> Beam:
    from ada import Placement
    from ada.core.vector_utils import transform3d

    logger.warning("Reading IFC Beams swept along IfcTrimmedCurve is WIP")

    r = axis.BasisCurve.Radius
    curve_place = get_placement(axis.BasisCurve.Position)
    beam_place = get_placement(ifc_elem.ObjectPlacement.RelativePlacement)
    p1 = get_point(axis.Trim1[1])
    p2 = get_point(axis.Trim2[1])
    global_place = Placement()
    angle = axis.Trim2[0].wrappedValue
    rot_origin = transform3d(beam_place.csys, global_place.csys, global_place.origin, [curve_place.origin])[0]
    rot_axis = transform3d(curve_place.csys, global_place.csys, global_place.origin, [curve_place.zdir])[0]

    p1g, p2g = transform3d(beam_place.csys, global_place.csys, beam_place.origin, [p1, p2])

    curve = CurveRevolve(p1g, p2g, radius=r, rot_axis=rot_axis, rot_origin=rot_origin, angle=np.rad2deg(angle))

    return Beam(
        name, curve=curve, sec=sec, mat=mat, guid=ifc_elem.GlobalId, ifc_store=ifc_store, units=ifc_store.assembly.units
    )


def import_polyline_beam(ifc_elem, axis, name, sec, mat, ifc_store: IfcStore) -> Beam:
    raise NotImplementedError("Reading beams swept along IfcPolyLines of length > 2 is not yet supported")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np

from ada.config import get_logger

from .transforms import Placement

if TYPE_CHECKING:
    from ada import FEM, Node
    from ada.fem import Surface

    from .primitives import PrimBox
    from .stru_beams import Beam
    from .stru_plates import Plate

logger = get_logger()


@dataclass
class BoundingBox:
    parent: PrimBox | Beam | Plate
    placement: Placement = field(default=None, init=False)
    sides: BoxSides = field(default=None, init=False)
    p1: np.array = field(default=None, init=False)
    p2: np.array = field(default=None, init=False)

    def __post_init__(self):
        from .primitives import Shape
        from .stru_beams import Beam
        from .stru_plates import Plate

        if issubclass(type(self.parent), Shape):
            self.p1, self.p2 = self._calc_bbox_of_shape()
            self.placement = self.parent.placement
        elif isinstance(self.parent, Beam):
            self.p1, self.p2 = self._calc_bbox_of_beam()
            self.placement = Placement(
                self.parent.placement.origin, xdir=self.parent.yvec, ydir=self.parent.xvec, zdir=self.parent.up
            )
        elif isinstance(self.parent, Plate):
            self.p1, self.p2 = self._calc_bbox_of_plate()
        else:
            raise NotImplementedError(f'Bounding Box Support for object type "{type(self.parent)}" is not yet added')
        self.sides = BoxSides(self)

    def _calc_bbox_of_beam(self) -> tuple[tuple, tuple]:
        """Get the bounding box of a beam"""
        from itertools import chain

        from ada import Beam, Section
        from ada.core.utils import roundoff
        from ada.sections.categories import BaseTypes

        bm = self.parent
        if bm.section.type == BaseTypes.CIRCULAR or bm.section.type == BaseTypes.TUBULAR:
            d = bm.section.r * 2
            dummy_beam = Beam("dummy", bm.n1.p, bm.n2.p, Section("DummySec", "BG", h=d, w_btn=d, w_top=d))
            outer_curve = dummy_beam.get_outer_points()
        else:
            outer_curve = bm.get_outer_points()

        points = np.array(list(chain.from_iterable(outer_curve)))
        xv = sorted([roundoff(p[0]) for p in points])
        yv = sorted([roundoff(p[1]) for p in points])
        zv = sorted([roundoff(p[2]) for p in points])
        xmin, xmax = xv[0], xv[-1]
        ymin, ymax = yv[0], yv[-1]
        zmin, zmax = zv[0], zv[-1]
        return (xmin, ymin, zmin), (xmax, ymax, zmax)

    def _calc_bbox_of_shape(self) -> tuple[tuple, tuple]:
        from .exceptions import NoGeomPassedToShapeError
        from .primitives import PrimBox

        if type(self.parent) is PrimBox:
            return self.parent.p1, self.parent.p2
        else:
            from ada.occ.utils import get_boundingbox

            try:
                return get_boundingbox(self.parent.geom(), use_mesh=True)
            except NoGeomPassedToShapeError as e:
                logger.info(f'Shape "{self.parent.name}" has no attached geometry. Error "{e}"')
                return (0, 0, 0), (1, 1, 1)

    def _calc_bbox_of_plate(self) -> tuple[tuple, tuple]:
        """Calculate the Bounding Box of a plate"""
        xs = []
        ys = []
        zs = []
        plate: Plate = self.parent
        for pt in plate.poly.nodes:
            xs.append(pt.x)
            ys.append(pt.y)
            zs.append(pt.z)

        bbox_min = np.array([min(xs), min(ys), min(zs)]).astype(np.float64)
        bbox_max = np.array([max(xs), max(ys), max(zs)]).astype(np.float64)
        n = plate.poly.normal.astype(np.float64)

        pv = np.nonzero(n)[0]
        matr = {0: "X", 1: "Y", 2: "Z"}
        orient = matr[pv[0]]
        if orient == "X" or orient == "Y":
            delta_vec = abs(n * plate.t / 2.0)
            bbox_min -= delta_vec
            bbox_max += delta_vec
        elif orient == "Z":
            delta_vec = abs(n * plate.t).astype(np.float64)
            bbox_min -= delta_vec
        else:
            raise ValueError(f"Error in {orient}")

        return tuple(bbox_min), tuple(bbox_max)

    @property
    def minmax(self):
        return self.p1, self.p2

    @property
    def volume_cog(self):
        """Get volumetric COG from bounding box"""

        return np.array(
            [
                (self.p1[0] + self.p2[0]) / 2,
                (self.p1[1] + self.p2[1]) / 2,
                (self.p1[2] + self.p2[2]) / 2,
            ]
        )


@dataclass
class BoxSides:
    parent: BoundingBox

    def _return_fem_nodes(self, pmin, pmax, fem):
        return fem.nodes.get_by_volume(p=pmin, vol_box=pmax)

    def _return_data(
        self, pmin, pmax, fem, return_fem_nodes, return_surface, surface_name, shell_positive
    ) -> tuple[tuple, tuple] | list[Node] | Surface:
        if return_fem_nodes is True or return_surface is True:
            part = self.parent.parent.parent
            if fem is None and self.parent is not None and part.fem.is_empty() is False:
                fem = part.fem

            if fem is None:
                raise ValueError("No FEM data found. Cannot return FEM nodes")

        if return_fem_nodes is True:
            return self._return_fem_nodes(pmin, pmax, fem)

        if return_surface is True:
            if surface_name is None:
                from .exceptions import NameIsNoneError

                raise NameIsNoneError("You must give 'surface_name' a string name unequal to None")
            nodes = self._return_fem_nodes(pmin, pmax, fem)
            if len(nodes) == 0:
                raise ValueError(f"Zero nodes found for (pmin, pmax): ({pmin}, {pmax})")
            return self._return_surface(surface_name, nodes, fem, shell_positive)
        return pmin, pmax

    def _return_surface(self, surface_name: str, nodes: list[Node], fem: FEM, shell_positive):
        from ada.fem.surfaces import create_surface_from_nodes

        return create_surface_from_nodes(surface_name, nodes, fem, shell_positive)

    def _get_dim(self):
        from ada import Beam

        bbox = self.parent
        p1 = np.array(bbox.p1)
        p2 = np.array(bbox.p2)

        bounded_obj = bbox.parent

        if type(bounded_obj) is Beam:
            l = bounded_obj.length
            w = max(bounded_obj.section.w_btn, bounded_obj.section.w_top)
            h = bounded_obj.section.h
        else:
            l, w, h = p2 - p1

        return l, w, h, p1, p2

    def top(
        self, tol=1e-3, return_fem_nodes=False, fem=None, return_surface=False, surf_name=None, surf_positive=False
    ):
        """Top is at positive local Z"""
        l, w, h, p1, p2 = self._get_dim()

        z = self.parent.placement.zdir

        pmin = p1 + h * z - tol
        pmax = p2 + tol

        return self._return_data(pmin, pmax, fem, return_fem_nodes, return_surface, surf_name, surf_positive)

    def bottom(
        self, tol=1e-3, return_fem_nodes=False, fem=None, return_surface=False, surface_name=None, surf_positive=False
    ):
        """Bottom is at negative local z"""
        l, w, h, p1, p2 = self._get_dim()

        z = self.parent.placement.zdir

        pmin = p1 - tol
        pmax = p2 - l * z + tol

        return self._return_data(pmin, pmax, fem, return_fem_nodes, return_surface, surface_name, surf_positive)

    def front(
        self, tol=1e-3, return_fem_nodes=False, fem=None, return_surface=False, surface_name=None, surf_positive=False
    ):
        """Front is at positive local y"""
        l, w, h, p1, p2 = self._get_dim()

        y = self.parent.placement.ydir

        pmin = p1 + l * y - tol
        pmax = p2 + tol

        return self._return_data(pmin, pmax, fem, return_fem_nodes, return_surface, surface_name, surf_positive)

    def back(
        self, tol=1e-3, return_fem_nodes=False, fem=None, return_surface=False, surface_name=None, surf_positive=False
    ):
        """Back is at negative local y"""
        l, w, h, p1, p2 = self._get_dim()

        y = self.parent.placement.ydir

        pmin = p1 - tol
        pmax = p2 - l * y + tol

        return self._return_data(pmin, pmax, fem, return_fem_nodes, return_surface, surface_name, surf_positive)



import xml.etree.ElementTree as ET
from typing import Dict, Union

from ada import Beam, Group, Part
from ada.config import get_logger

logger = get_logger()


def get_sets(xml_root: ET.Element, parent: Part) -> Dict[str, Group]:
    el_sets = dict()
    for el_set in xml_root.findall(".//set"):
        name = el_set.attrib["name"]
        members = list(filter(lambda x: x is not None, [get_concept(m, parent) for m in el_set.findall(".//concept")]))
        el_sets[name] = Group(name, members, parent=parent)

    return el_sets


def get_concept(xml_el: ET.Element, part: Part) -> Union[Beam]:
    ref = xml_el.attrib["concept_ref"]
    if ref in part.beams.dmap.keys():
        return part.beams.from_name(ref)
    else:
        logger.debug(f'Currently only Beams are supported. Unable to find group member "{ref}"')



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada.fem.formats.general import FEATypes
    from ada.fem.results.common import FEAResult


def postprocess(res_path: str | pathlib.Path, fem_format: FEATypes = None) -> FEAResult:
    from ada.fem.formats.abaqus.results.read_odb import read_odb_pckle_file
    from ada.fem.formats.calculix.results.read_frd_file import read_from_frd_file_proto
    from ada.fem.formats.code_aster.results.read_rmed_results import read_rmed_file
    from ada.fem.formats.general import FEATypes
    from ada.fem.formats.sesam.results.read_sif import read_sif_file
    from ada.fem.formats.utils import interpret_fem_format_from_path

    if fem_format is None:
        fem_format = interpret_fem_format_from_path(res_path)

    if isinstance(fem_format, str):
        fem_format = FEATypes.from_str(fem_format)

    if fem_format == FEATypes.SESAM:
        return read_sif_file(res_path)
    elif fem_format == FEATypes.ABAQUS:
        return read_odb_pckle_file(res_path)
    elif fem_format == FEATypes.CALCULIX:
        return read_from_frd_file_proto(res_path)
    elif fem_format == FEATypes.CODE_ASTER:
        return read_rmed_file(res_path)
    else:
        raise NotImplementedError()



from typing import TYPE_CHECKING

from ..utils import create_guid, create_ifc_placement, ifc_dir, ifc_p, to_real

if TYPE_CHECKING:
    from ifcopenshell import file

    from ada.concepts.transforms import Instance


def write_mapped_instance(instance: "Instance", f: "file"):
    elem = instance.instance_ref
    products = list(filter(lambda x: x.GlobalId == elem.guid, f.by_type("IfcProduct")))
    if len(products) != 1:
        raise ValueError(f'Unable to find IFC element with guid="{elem.guid}"')
    ifc_elem = products[0]
    origin = create_ifc_placement(f)  # , loc_z=elem.xvec.astype(float).tolist()
    body = ifc_elem.Representation.Representations[0]

    rep_map = f.create_entity("IFCREPRESENTATIONMAP", origin, body)
    mapped_instances = []
    for place in instance.placements:
        tra = f.create_entity(
            "IFCCARTESIANTRANSFORMATIONOPERATOR3DNONUNIFORM",
            Axis1=ifc_dir(f, place.xdir),
            Axis2=ifc_dir(f, place.ydir),
            LocalOrigin=ifc_p(f, place.origin),
            Scale=to_real(place.scale),
            Axis3=ifc_dir(f, place.zdir),
            Scale2=to_real(place.scale),
            Scale3=to_real(place.scale),
        )

        mapped_item = f.create_entity("IFCMAPPEDITEM", rep_map, tra)
        mapped_instances.append(mapped_item)

    shape_rep = f.create_entity(
        "IFCSHAPEREPRESENTATION", body.ContextOfItems, "body", "MappedRepresentation", mapped_instances
    )

    prod_def_shape = f.create_entity(
        "IFCPRODUCTDEFINITIONSHAPE", Name=None, Description=None, Representations=[shape_rep]
    )

    f.create_entity(
        "IFCBUILDINGELEMENTPROXY",
        create_guid(),
        None,
        elem.name + "_instances",
        "Mapped Instances",
        None,
        prod_def_shape,
        None,
        None,
    )

    # 32= IFCBUILDINGELEMENTPROXY('1kTvXnbbzCWw8lcMd1dR4o',$,'P-1','sample proxy',$,#44,#24,$,$);



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import Section


class BaseTypes(Enum):
    BOX = "BOX"
    TUBULAR = "TUB"
    IPROFILE = "I"
    TPROFILE = "T"
    ANGULAR = "HP"
    CHANNEL = "UNP"
    CIRCULAR = "CIRC"
    GENERAL = "GENERAL"
    FLATBAR = "FB"
    POLY = "poly"

    @staticmethod
    def from_str(type_str: str):
        key_map = {x.value.lower(): x for x in BaseTypes}
        result = key_map.get(type_str.lower(), None)

        if result is None:
            result = SectionCat.get_shape_type(type_str.upper())
            if result is None:
                raise ValueError()

        return result


class SectionCat:
    BASETYPES = BaseTypes

    box = [BASETYPES.BOX.value, "BG", "CG"]
    shs = ["SHS"]
    rhs = ["RHS", "URHS"]
    tubular = [BASETYPES.TUBULAR.value, "PIPE", "OD"]
    iprofiles = ["HEA", "HEB", "HEM", "IPE"]
    igirders = [BASETYPES.IPROFILE.value, "IG"]
    tprofiles = [BASETYPES.TPROFILE.value, "TG"]
    angular = [BASETYPES.ANGULAR.value]
    channels = [BASETYPES.CHANNEL.value]
    circular = [BASETYPES.CIRCULAR.value]
    general = [BASETYPES.GENERAL.value, "GENBEAM"]
    flatbar = [BASETYPES.FLATBAR.value]
    poly = ["POLY"]

    @staticmethod
    def _get_sec_type(section_ref: str | Section) -> str | BaseTypes:
        from ada import Section

        if isinstance(section_ref, Section):
            return section_ref.type
        return section_ref.upper()

    @classmethod
    def get_shape_type(cls, bm_type):
        type_map = [
            (cls.is_i_profile, cls.BASETYPES.IPROFILE),
            (cls.is_angular, cls.BASETYPES.ANGULAR),
            (cls.is_flatbar, cls.BASETYPES.FLATBAR),
            (cls.is_box_profile, cls.BASETYPES.BOX),
            (cls.is_t_profile, cls.BASETYPES.TPROFILE),
            (cls.is_channel_profile, cls.BASETYPES.CHANNEL),
            (cls.is_tubular_profile, cls.BASETYPES.TUBULAR),
            (cls.is_circular_profile, cls.BASETYPES.CIRCULAR),
            (cls.is_general, cls.BASETYPES.GENERAL),
            (cls.is_poly, cls.BASETYPES.POLY),
        ]

        for type_func, return_type in type_map:
            if type_func(bm_type):
                return return_type

    @classmethod
    def is_i_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.igirders + cls.iprofiles

    @classmethod
    def is_t_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.tprofiles

    @classmethod
    def is_box_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.box + cls.shs + cls.rhs

    @classmethod
    def is_circular_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.circular

    @classmethod
    def is_tubular_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.tubular

    @classmethod
    def is_channel_profile(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.channels

    @classmethod
    def is_flatbar(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.flatbar

    @classmethod
    def is_general(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.general

    @classmethod
    def is_angular(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.angular

    @classmethod
    def is_poly(cls, bmtype) -> bool:
        return cls._get_sec_type(bmtype) in cls.poly

    @classmethod
    def is_strong_axis_symmetric(cls, section: Section) -> bool:
        return section.w_top == section.w_btn and section.t_ftop == section.t_fbtn



import json

import h5py

from ada import Beam, Material, Section
from ada.concepts.containers import Beams, Materials, Nodes, Sections
from ada.concepts.points import Node
from ada.concepts.spatial import Assembly, Part
from ada.config import get_logger
from ada.fem import FEM, Elem
from ada.fem.containers import FemElements
from ada.materials.metals import CarbonSteel

from .utils import from_safe_name, str_fix

logger = get_logger()


def read_assembly_from_cache(h5_filename, assembly=None):
    import h5py

    with h5py.File(h5_filename, "r") as f:
        info = f["INFO"].attrs
        a = Assembly(info["NAME"]) if assembly is None else assembly
        walk_parts(f.get("PARTS"), a)

    return a


def walk_parts(cache_p, parent):
    for name, p in cache_p.items():
        if type(p) is h5py.Dataset:
            continue
        unsafe_name = from_safe_name(name)
        parent_name = from_safe_name(p.attrs.get("PARENT", ""))
        if parent.name != parent_name:
            logger.error("Unable to retrieve proper Hierarchy from HDF5 cache")
        curr_p = parent.add_part(get_part_from_cache(unsafe_name, p))
        walk_parts(p, curr_p)


def get_part_from_cache(name, part_cache):
    meta_str = part_cache.attrs.get("METADATA")
    metadata = None
    if meta_str is not None:
        metadata = json.loads(meta_str)

    p = Part(name, metadata=metadata)
    fem = part_cache.get("FEM")
    if fem is not None:
        p._fem = get_fem_from_cache(fem)

    node_group = part_cache.get("NODES")
    if node_group is not None:
        p._nodes = get_nodes_from_cache(node_group, p)

    sections = get_sections_from_cache(part_cache, p)
    if sections is not None:
        p._sections = sections

    materials = get_materials_from_cache(part_cache, p)
    if materials is not None:
        p._materials = materials

    beams = get_beams_from_cache(part_cache, p)
    if beams is not None:
        p._beams = beams

    return p


def get_beams_from_cache(part_cache, parent: Part):
    prefix = "BEAMS"
    beams_str = part_cache.get(f"{prefix}_STR")
    beams_int = part_cache.get(f"{prefix}_INT")
    beams_up = part_cache.get(f"{prefix}_UP")
    if beams_str is None:
        return None

    def bm_from_cache(bm_str, bm_int, bm_up):
        nid1, nid2 = [parent.nodes.from_id(nid) for nid in bm_int]
        guid, name, sec_name, mat_name, meta_str = str_fix(bm_str)
        sec = parent.sections.get_by_name(sec_name)
        mat = parent.materials.get_by_name(mat_name)
        metadata = None
        if meta_str is not None:
            metadata = json.loads(meta_str)
        return Beam(name, nid1, nid2, sec=sec, mat=mat, guid=guid, parent=parent, metadata=metadata, up=bm_up)

    bm_zip = zip(beams_str, beams_int, beams_up)

    return Beams([bm_from_cache(bm_str, bm_int, bm_up) for bm_str, bm_int, bm_up in bm_zip], parent=parent)


def get_sections_from_cache(part_cache, parent):
    sections_str = part_cache.get("SECTIONS_STR")
    sections_int = part_cache.get("SECTIONS_INT")
    if sections_str is None:
        return None

    def sec_from_list(sec_str, sec_int):
        guid, name, units, sec_type = str_fix(sec_str)
        r, wt, h, w_top, w_btn, t_w, t_ftop, t_fbtn, sec_id = [x if x != 0 else None for x in sec_int]
        return Section(
            name=name,
            guid=guid,
            sec_id=sec_id,
            units=units,
            sec_type=sec_type,
            r=r,
            wt=wt,
            h=h,
            w_top=w_top,
            w_btn=w_btn,
            t_w=t_w,
            t_ftop=t_ftop,
            t_fbtn=t_fbtn,
        )

    return Sections(
        [sec_from_list(sec_str, sec_int) for sec_str, sec_int in zip(sections_str, sections_int)], parent=parent
    )


def get_materials_from_cache(part_cache, parent):
    mat_str = part_cache.get("MATERIALS_STR")
    mat_int = part_cache.get("MATERIALS_INT")

    if mat_str is None:
        return None

    def mat_from_list(mat_int, mat_str):
        guid, name, units = str_fix(mat_str)
        E, rho, sigy, mat_id = mat_int
        return Material(
            name=name,
            guid=guid,
            mat_id=mat_id,
            units=units,
            mat_model=CarbonSteel(E=E, rho=rho, sig_y=sigy),
            parent=parent,
        )

    return Materials([mat_from_list(mat_int, mat_str) for mat_int, mat_str in zip(mat_int, mat_str)], parent=parent)


def get_fem_from_cache(cache_fem):
    node_groups = cache_fem["NODES"]
    fem = FEM(cache_fem.attrs["NAME"])
    fem.nodes = get_nodes_from_cache(node_groups, fem)
    elements = []
    for eltype, mesh in cache_fem["MESH"].items():
        el_ids = mesh["ELEMENTS"][()]
        elements += [Elem(el_id[0], [fem.nodes.from_id(eli) for eli in el_id[1:]], eltype) for el_id in el_ids]
    fem.elements = FemElements(elements, fem)

    return fem


def get_nodes_from_cache(node_group, parent):
    points_in = node_group[()]
    points = [Node(n[1:], n[0]) for n in points_in]
    return Nodes(points, parent=parent)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem import Spring


def springs_str(fem: FEM):
    return "\n".join([spring_str(c) for c in fem.springs.values()]) if len(fem.springs) > 0 else "** No Springs"


def spring_str(spring: Spring) -> str:
    from ada.fem.shapes.definitions import SpringTypes

    if spring.type not in (SpringTypes.SPRING1,):
        raise ValueError(f'Currently unsupported spring type "{spring.type}"')

    _str = f'** Spring El "{spring.name}"\n\n'
    for dof, row in enumerate(spring.stiff):
        for j, stiffness in enumerate(row):
            if dof == j:
                _str += f"""*Spring, elset={spring.fem_set.name}
 {dof + 1}
 {stiffness:.6E}
{spring.id}, {spring.nodes[0].id}\n"""
    return _str.rstrip()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from ada import Beam, CurvePoly, CurveRevolve
from ada.config import Settings, get_logger
from ada.core.constants import O
from ada.ifc.utils import (
    add_colour,
    convert_bm_jusl_to_ifc,
    create_guid,
    create_ifc_placement,
    create_local_placement,
    ifc_dir,
    ifc_p,
    to_real,
)
from ada.ifc.write.write_curves import write_curve_poly

if TYPE_CHECKING:
    from ifcopenshell import file as ifile

    from ada.ifc.store import IfcStore

logger = get_logger()


def write_ifc_beam(ifc_store: IfcStore, beam: Beam):
    ibw = IfcBeamWriter(ifc_store)
    return ibw.create_ifc_beam(beam)


@dataclass
class IfcBeamWriter:
    ifc_store: IfcStore

    def create_ifc_beam(self, beam: Beam):
        if beam.parent is None:
            raise ValueError("Parent cannot be None for IFC export")

        f = self.ifc_store.f

        owner_history = self.ifc_store.owner_history
        profile = self.ifc_store.get_profile_def(beam.section)

        if isinstance(beam.curve, CurveRevolve):
            axis, body, loc_plac = create_revolved_beam(beam, f, profile)
        elif isinstance(beam.curve, CurvePoly):
            axis, body, loc_plac = create_polyline_beam(beam, f, profile)
        else:
            if beam.curve is not None:
                raise ValueError(f'Unrecognized beam.curve "{type(beam.curve)}"')
            axis, body, loc_plac = extrude_straight_beam(beam, f, profile)

        prod_def_shp = f.create_entity("IfcProductDefinitionShape", None, None, (axis, body))

        ifc_beam = f.create_entity(
            "IfcBeam",
            GlobalId=beam.guid,
            OwnerHistory=owner_history,
            Name=beam.name,
            Description=beam.section.sec_str,
            ObjectType="Beam",
            ObjectPlacement=loc_plac,
            Representation=prod_def_shp,
        )

        found_existing_relationship = False

        beam_type = self.ifc_store.get_beam_type(beam.section)
        if beam_type is None:
            raise ValueError()

        for ifcrel in f.by_type("IfcRelDefinesByType"):
            if ifcrel.RelatingType == beam_type:
                ifcrel.RelatedObjects = tuple([*ifcrel.RelatedObjects, ifc_beam])
                found_existing_relationship = True
                break

        if found_existing_relationship is False:
            f.create_entity(
                "IfcRelDefinesByType",
                GlobalId=create_guid(),
                OwnerHistory=owner_history,
                Name=beam.section.type.value,
                Description=None,
                RelatedObjects=[ifc_beam],
                RelatingType=beam_type,
            )

        self.add_material_assignment(beam, ifc_beam)

        return ifc_beam

    def add_material_assignment(self, beam: Beam, ifc_beam):
        sec = beam.section
        mat = beam.material
        ifc_store = self.ifc_store
        f = ifc_store.f

        ifc_mat_rel = ifc_store.f.by_guid(mat.guid)
        ifc_mat = ifc_mat_rel.RelatingMaterial

        ifc_profile = ifc_store.get_profile_def(beam.section)
        mat_profile = f.createIfcMaterialProfile(
            sec.name, "A material profile", ifc_mat, ifc_profile, None, "LoadBearing"
        )
        mat_profile_set = f.createIfcMaterialProfileSet(sec.name, None, [mat_profile], None)

        mat_usage = f.create_entity("IfcMaterialProfileSetUsage", mat_profile_set, convert_bm_jusl_to_ifc(beam))

        ifc_store.writer.create_rel_associates_material(create_guid(), mat_usage, [ifc_beam])
        ifc_store.writer.associate_elem_with_material(beam.material, ifc_beam)

        return mat_profile_set


def extrude_straight_beam(beam: Beam, f: ifile, profile):
    extrude_dir = ifc_dir(f, (0.0, 0.0, 1.0))
    parent = f.by_guid(beam.parent.guid)
    a = beam.parent.get_assembly()

    global_placement = create_local_placement(f, relative_to=parent.ObjectPlacement)

    e1 = (0.0, 0.0, 0.0)

    vec = beam.xvec
    yvec = beam.yvec
    if Settings.model_export.include_ecc and beam.e1 is not None:
        e1 = beam.e1
        vec = beam.xvec_e

    profile_e = None
    if beam.taper is not None and beam.section != beam.taper:
        profile_e = f.by_guid(beam.taper.guid)

    # Transform coordinates to local coords
    p1 = tuple([float(x) + float(e1[i]) for i, x in enumerate(beam.n1.p.copy())])
    p2 = p1 + np.array([0, 0, 1]) * beam.length

    p1_ifc = f.create_entity("IfcCartesianPoint", to_real(p1))
    p2_ifc = f.create_entity("IfcCartesianPoint", to_real(p2))

    ifc_polyline = f.create_entity("IfcPolyLine", [p1_ifc, p2_ifc])

    global_origin = f.createIfcCartesianPoint(O)
    ifc_axis2plac3d = f.create_entity("IfcAxis2Placement3D", global_origin, None, None)

    if profile_e is not None:
        extrude_area_solid = f.create_entity(
            "IfcExtrudedAreaSolidTapered", profile, ifc_axis2plac3d, extrude_dir, beam.length, profile_e
        )
    else:
        extrude_area_solid = f.create_entity("IfcExtrudedAreaSolid", profile, ifc_axis2plac3d, extrude_dir, beam.length)

    # Add colour
    if beam.colour is not None:
        add_colour(f, extrude_area_solid, str(beam.colour), beam.colour)

    body_context = a.ifc_store.get_context("Body")
    axis_context = a.ifc_store.get_context("Axis")
    ax23d = f.create_entity("IfcAxis2Placement3D", p1_ifc, ifc_dir(f, vec), ifc_dir(f, yvec))
    loc_plac = f.create_entity("IfcLocalPlacement", global_placement, ax23d)
    body = f.create_entity("IfcShapeRepresentation", body_context, "Body", "SweptSolid", [extrude_area_solid])
    axis = f.create_entity("IfcShapeRepresentation", axis_context, "Axis", "Curve3D", [ifc_polyline])

    return body, axis, loc_plac


def create_revolved_beam(beam: Beam, f: "ifile", profile):
    a = beam.parent.get_assembly()
    body_context = a.ifc_store.get_context("Body")
    axis_context = a.ifc_store.get_context("Axis")

    curve: CurveRevolve = beam.curve

    ifc_trim_curve = create_ifc_trimmed_curve(curve, f)
    placement = create_local_placement(f, curve.p1, (0, 0, 1))
    solid = create_ifcrevolveareasolid(f, profile, placement, curve.p1, curve.rot_axis, np.deg2rad(curve.angle))

    axis = f.create_entity("IfcShapeRepresentation", axis_context, "Axis", "Curve3D", [ifc_trim_curve])
    body = f.create_entity("IfcShapeRepresentation", body_context, "Body", "SweptSolid", [solid])

    return body, axis, placement


def create_ifc_trimmed_curve(curve: CurveRevolve, f: "ifile"):
    loc_plac = create_ifc_placement(f, origin=curve.rot_origin)
    ifc_circle = f.create_entity("IFCCIRCLE", loc_plac, curve.radius)
    param1 = (f.create_entity("IFCPARAMETERVALUE", 0.0), ifc_p(f, curve.p1))
    param2 = (f.create_entity("IFCPARAMETERVALUE", np.deg2rad(curve.angle)), ifc_p(f, curve.p2))
    trim_curve = f.create_entity(
        "IFCTRIMMEDCURVE",
        BasisCurve=ifc_circle,
        Trim1=param1,
        Trim2=param2,
        SenseAgreement=True,
        MasterRepresentation="PARAMETER",
    )
    return trim_curve


def create_ifcrevolveareasolid(f, profile, ifcaxis2placement, origin, revolve_axis, revolve_angle):
    """Creates an IfcExtrudedAreaSolid from a list of points, specified as Python tuples"""
    ifcaxis1dir = f.create_entity("IfcAxis1Placement", ifc_p(f, origin), ifc_dir(f, revolve_axis))
    return f.create_entity("IfcRevolvedAreaSolid", profile, ifcaxis2placement, ifcaxis1dir, revolve_angle)


def create_polyline_beam(beam, f, profile):
    ifc_polyline = write_curve_poly(beam.curve)

    extrude_dir = ifc_dir(f, (0.0, 0.0, 1.0))
    global_placement = create_ifc_placement(f)

    extrude_area_solid = f.create_entity(
        "IfcFixedReferenceSweptAreaSolid", profile, global_placement, ifc_polyline, 0.0, 1.0, extrude_dir
    )
    loc_plac = create_ifc_placement(f)
    return extrude_area_solid, loc_plac, ifc_polyline


def sweep_beam(beam, f, profile, global_placement, extrude_dir):
    ifc_polyline = write_curve_poly(beam.curve)

    extrude_area_solid = f.create_entity(
        "IfcFixedReferenceSweptAreaSolid", profile, global_placement, ifc_polyline, 0.0, 1.0, extrude_dir
    )
    loc_plac = create_ifc_placement(f)
    return extrude_area_solid, loc_plac, ifc_polyline



# ABAQUS/PYTHON POST PROCESSING SCRIPT
# Run using abaqus python / abaqus viewer -noGUI / abaqus cae -noGUI
# Tip! Import this into Abaqus PDE directly to further develop this script.
import inspect
import logging
import os
import sys
import traceback

import cPickle as pickle
import numpy as np
import symbolicConstants
from odbAccess import (
    ELEMENT_NODAL,
    INTEGRATION_POINT,
    NODAL,
    OdbAssemblyType,
    OdbInstanceType,
    OdbMeshElementArrayType,
    OdbMeshNodeArrayType,
    OdbStepType,
    openOdb,
)
from symbolicConstants import SymbolicConstant

_constants = [x[1] for x in inspect.getmembers(symbolicConstants, inspect.isclass)]
logger = logging.getLogger("abaqus")


def filter1(obj, attr):
    if attr[:1] == "_":
        return False
    if hasattr(obj, attr) is False:
        return False
    r = getattr(obj, attr)
    if callable(r):
        return False
    return True


def is_constant(value):
    if value in _constants:
        return True
    if isinstance(value, SymbolicConstant):
        return True
    return False


def get_data_from_attr(obj, attributes):
    att_filter = filter(lambda attr: filter1(obj, attr), attributes)
    return {attr: serialize(getattr(obj, attr)) for attr in att_filter}


instance_names = []


def serialize(obj):
    """Recursively walk object's hierarchy."""
    if isinstance(obj, (int, float)):
        return obj
    elif isinstance(obj, np.float32):
        return float(obj)
    elif isinstance(obj, (list, tuple)):
        return [serialize(item) for item in obj]
    elif isinstance(obj, (np.ndarray,)):
        return [serialize(v) for v in obj]
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, dict):
        return {key: serialize(val) for key, val in obj.items()}
    elif isinstance(obj, object):
        attributes = dir(obj)
        if "values" in attributes:
            return serialize({key: serialize(value) for key, value in obj.items()})
        else:
            if isinstance(obj, OdbAssemblyType):
                data = get_data_from_attr(obj, attributes)
                data["instances"] = [instance_data(x) for x in obj.instances.values()]
                return data
            if isinstance(obj, OdbInstanceType):
                return obj.name
            if isinstance(obj, OdbMeshElementArrayType):
                return [x.label for x in obj]
            if isinstance(obj, OdbMeshNodeArrayType):
                return [x.label for x in obj]
            if isinstance(obj, OdbStepType):
                data = get_data_from_attr(obj, attributes)
                # frames does not get exported automatically
                data["frames"] = [get_frame_data(frame) for frame in obj.frames]
                return data
            if is_constant(obj):
                return str(obj)

            return get_data_from_attr(obj, attributes)
    else:
        raise ValueError('Unknown entity "{}", "{}"'.format(type(obj), obj))


def instance_data(obj):
    data = dict(name=obj.name)

    sc_ = {el.sectionCategory.name: el.sectionCategory for el in obj.elements}
    sc = {key: (i, [serialize(x) for x in value.sectionPoints]) for i, (key, value) in enumerate(sc_.items())}
    data["section_cats"] = sc

    data["nodeSets"] = serialize(obj.nodeSets)
    data["elementSets"] = serialize(obj.elementSets)
    data["beamOrientations"] = [x.vector for x in obj.beamOrientations] if obj.beamOrientations is not None else None
    data["nodes"] = [(n.label, [float(x) for x in n.coordinates]) for n in obj.nodes]
    data["elements"] = [(el.label, el.type, el.connectivity, sc.get(el.sectionCategory.name)[0]) for el in obj.elements]

    return data


def get_nodal_value(n):
    node_label = int(n.nodeLabel)
    element_label = int(n.elementLabel)
    sec_p = int(n.sectionPoint.number) if n.sectionPoint is not None else None
    return dict(nodeLabel=node_label, elementLabel=element_label, sec_p_num=sec_p, data=serialize(n.data))


def get_field_data(field):
    num_locs = len(field.locations)
    if num_locs > 1:
        logger.info("FieldOutput {} contains multiple section locations".format(field.name))
    curr_pos = field.locations[0].position
    if curr_pos == INTEGRATION_POINT:  # values obtained by extrapolating results calculated at the integration points.
        nodal_data = field.getSubset(position=ELEMENT_NODAL)
        components = nodal_data.componentLabels
        return "ELEMENT_NODAL", components, [get_nodal_value(n) for n in nodal_data.values]
    if curr_pos == NODAL:  # specifying the values calculated at the nodes.
        nodal_data = field.getSubset(position=NODAL)
        components = nodal_data.componentLabels
        return "NODAL", components, [(int(n.nodeLabel), serialize(n.data)) for n in nodal_data.values]

    logger.info("Skipping unsupported field position {}".format(curr_pos))
    return None


def get_frame_data(frame):
    data = dict()
    for key, field in frame.fieldOutputs.items():
        field_data = dict()
        field_values = get_field_data(field)
        if field_values is None:
            continue
        field_data["values"] = field_values
        data[key] = field_data

    return data


def get_section_data(section_obj):
    data = dict(name=section_obj.name)
    if hasattr(section_obj, "profile"):
        data["profile"] = serialize(getattr(section_obj, "profile"))
    elif hasattr(section_obj, "thickness"):
        data["thickness"] = serialize(getattr(section_obj, "thickness"))
    return data


analysis_path = sys.argv[1]
parent_dir = os.path.dirname(analysis_path)
logger.basicConfig(
    filename=os.path.join(parent_dir, "aba_io.log"),
    filemode="w",
    level=logger.INFO,
    format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger.info('Logging to "{}"'.format(parent_dir))
logger.info('Opening ODB "{}"'.format(analysis_path))

odb = openOdb(analysis_path, readOnly=True)

try:
    fname = os.path.basename(analysis_path)
    logger.info('serializing "{}" ODB data'.format(fname))
    res = serialize(
        dict(
            rootAssembly=odb.rootAssembly,
            sections={key: get_section_data(value) for key, value in odb.sections.items()},
            profiles={key: dict(name=value.__class__.__name__, value=value) for key, value in odb.profiles.items()},
            userdata=odb.userData,
            steps=odb.steps,
            last_updated=os.path.getmtime(analysis_path),
        )
    )
    logger.info("serialization complete")

    logger.info("Starting export to pickle")
    pckle_name = fname.replace(".odb", ".pckle")
    with open(os.path.join(parent_dir, pckle_name), "wb") as f:
        pickle.dump(res, f, 2)

except (BaseException, RuntimeError) as e:
    trace_str = traceback.format_exc()
    logger.error("{}, {}".format(e, trace_str))
finally:
    odb.close()

logger.info("Export to pickle complete")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


def resolve_ids_in_multiple(tags, tags_data, is_elem):
    """Find elements shared by multiple sets"""
    from ada.fem import FemSet

    rmap = {tuple(v): r for r, v in tags.items()}
    fin_data = dict()
    for t, memb in tags_data.items():
        fin_data[t] = []
        for mem in memb:
            refs = list(filter(lambda x: type(x) == FemSet, mem.refs))
            if len(refs) > 1:
                names = [r.name for r in refs]
                if names not in tags.values():
                    new_int = min(tags.keys()) - 1 if is_elem else max(tags.keys()) + 1
                    tags[new_int] = names
                    rmap[tuple(names)] = new_int
                    fin_data[new_int] = []
                else:
                    new_int = rmap[tuple(names)]
                if mem not in fin_data[new_int]:
                    fin_data[new_int].append(mem)
            else:
                fin_data[t].append(mem)
    to_be_removed = []
    for i, f in fin_data.items():
        if len(f) == 0:
            to_be_removed.append(i)
    for t in to_be_removed:
        fin_data.pop(t)
        tags.pop(t)
    return fin_data



from operator import attrgetter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ada import FEM


def nodes_str(fem: "FEM"):
    f = "{nid:>7}, {x:>13.6f}, {y:>13.6f}, {z:>13.6f}"
    if len(fem.nodes) == 0:
        return "** No Nodes"
    return (
        "*NODE\n"
        + "\n".join(
            [f.format(nid=no.id, x=no[0], y=no[1], z=no[2]) for no in sorted(fem.nodes, key=attrgetter("id"))]
        ).rstrip()
    )


def rp_str(fem: "FEM") -> str:
    from .write_sets import aba_set_str

    f = "{nid:>7}, {x:>13.6f}, {y:>13.6f}, {z:>13.6f}"

    if len(fem.ref_points.nodes) == 0:
        return "** No Nodes"

    ref_int = fem.nodes.max_nid
    fem.ref_points.renumber(int(ref_int + 1))

    rp_nodes_str = (
        "*NODE\n"
        + "\n".join(
            [f.format(nid=no.id, x=no[0], y=no[1], z=no[2]) for no in sorted(fem.ref_points, key=attrgetter("id"))]
        ).rstrip()
    )
    for nset in fem.ref_sets:
        nset.name += "-RefPt_"
    rp_sets_str = "\n" + "\n".join([aba_set_str(no, True, False) for no in fem.ref_sets]).rstrip()

    return rp_nodes_str + rp_sets_str



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import h5py
import numpy as np

from ada.fem.formats.utils import get_fem_model_from_assembly
from ada.fem.utils import is_quad8_shell_elem, is_tri6_shell_elem

from ..compatibility import check_compatibility
from .templates import el_convert_str, main_comm_str
from .write_bc import create_bc_str
from .write_elements import elements_str
from .write_materials import materials_str
from .write_nodes import _write_nodes
from .write_sections import create_sections_str
from .write_steps import create_step_str

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly, Part


def to_fem(assembly: Assembly, name, analysis_dir, metadata=None):
    """Write Code_Aster .med and .comm file from Assembly data"""
    from ada.materials.utils import shorten_material_names

    check_compatibility(assembly)

    if "info" not in metadata:
        metadata["info"] = dict(description="")

    p = get_fem_model_from_assembly(assembly)
    # Prepare model for
    shorten_material_names(assembly)
    # TODO: Implement support for multiple parts. Need to understand how submeshes in Salome and Code Aster works.
    # for p in filter(lambda x: len(x.fem.elements) != 0, assembly.get_all_parts_in_assembly(True)):
    write_to_med(name, p, analysis_dir)

    with open((analysis_dir / name).with_suffix(".comm"), "w") as f:
        f.write(create_comm_str(assembly, p))

    print(f'Created a Code_Aster input deck at "{analysis_dir}"')


def create_comm_str(assembly: Assembly, part: Part) -> str:
    """Create COMM file input str"""
    mat_str = materials_str(assembly)
    sections_str = create_sections_str(part.fem.sections)
    bcs = part.fem.bcs
    if assembly != part:
        bcs += assembly.fem.bcs
    bc_str = "\n".join([create_bc_str(bc) for bc in bcs])
    step_str = "\n".join([create_step_str(s, part) for s in assembly.fem.steps])

    type_tmpl_str = "_F(GROUP_MA={elset_str}, PHENOMENE='MECANIQUE', MODELISATION='{el_formula}',),"

    model_type_str = ""
    section_sets = ""
    input_mesh = "mesh"
    if len(part.fem.sections.lines) > 0:
        bm_elset_str = ",".join([f"'{bm_fs.elset.name}'" for bm_fs in part.fem.sections.lines])
        section_sets += f"bm_sets = ({bm_elset_str})\n"
        model_type_str += type_tmpl_str.format(elset_str="bm_sets", el_formula="POU_D_E")

    if len(part.fem.sections.shells) > 0:
        sh_elset_str = ""
        second_order = ""
        is_tri6 = False
        is_quad8 = False
        for sh_fs in part.fem.sections.shells:
            is_quad8 = is_quad8_shell_elem(sh_fs)
            is_tri6 = is_tri6_shell_elem(sh_fs)
            if is_tri6 or is_quad8:
                second_order += f"'{sh_fs.elset.name}',"
            else:
                sh_elset_str += f"'{sh_fs.elset.name}',"

        if sh_elset_str != "":
            elset = "sh_sets"
            section_sets += f"{elset} = ({sh_elset_str})\n"
            model_type_str += type_tmpl_str.format(elset_str=elset, el_formula="DKT")

        if second_order != "":
            elset = "sh_2nd_order_sets"
            section_sets += f"{elset} = ({second_order})\n"

            if is_tri6:
                output_mesh = "ma_tri6"
                section_sets += el_convert_str.format(
                    output_mesh=output_mesh, input_mesh=input_mesh, el_set=elset, convert_option="TRIA6_7"
                )
                input_mesh = output_mesh

            if is_quad8:
                output_mesh = "ma_quad8"
                section_sets += el_convert_str.format(
                    output_mesh=output_mesh, input_mesh=input_mesh, el_set=elset, convert_option="QUAD8_9"
                )
                input_mesh = output_mesh
            model_type_str += type_tmpl_str.format(elset_str=elset, el_formula="COQUE_3D")

    if len(part.fem.sections.solids) > 0:
        so_elset_str = ",".join([f"'{solid_fs.elset.name}'" for solid_fs in part.fem.sections.solids])
        section_sets += f"so_sets = ({so_elset_str})\n"
        model_type_str += type_tmpl_str.format(elset_str="so_sets", el_formula="3D")

    comm_str = main_comm_str.format(
        section_sets=section_sets,
        input_mesh=input_mesh,
        model_type_str=model_type_str,
        materials_str=mat_str,
        sections_str=sections_str,
        bc_str=bc_str,
        step_str=step_str,
    )

    return comm_str


def write_to_med(name, part: Part, analysis_dir):
    """Custom Method for writing a part directly based on meshio"""

    analysis_dir = pathlib.Path(analysis_dir)
    filename = (analysis_dir / name).with_suffix(".med")

    with h5py.File(filename, "w") as f:
        mesh_name = name if name is not None else part.fem.name
        # Strangely the version must be 3.0.x
        # Any version >= 3.1.0 will NOT work with SALOME 8.3
        info = f.create_group("INFOS_GENERALES")
        info.attrs.create("MAJ", 3)
        info.attrs.create("MIN", 0)
        info.attrs.create("REL", 0)

        time_step = _write_mesh_presets(f, mesh_name)

        profile = "MED_NO_PROFILE_INTERNAL"

        # Node and Element sets (familles in French)
        fas = f.create_group("FAS")
        families = fas.create_group(mesh_name)
        family_zero = families.create_group("FAMILLE_ZERO")  # must be defined in any case
        family_zero.attrs.create("NUM", 0)

        # Make sure that all member references are updated
        # TODO: Evaluate if this can be avoided using a smarter algorithm
        part.fem.sets.add_references()

        # Nodes and node sets
        _write_nodes(part, time_step, profile, families)

        # Elements (mailles in French) and element sets
        elements_str(part, time_step, profile, families)


def _write_mesh_presets(f, mesh_name):
    numpy_void_str = np.string_("")
    dim = 3

    # Meshes
    mesh_ensemble = f.create_group("ENS_MAA")

    med_mesh = mesh_ensemble.create_group(mesh_name)
    med_mesh.attrs.create("DIM", dim)  # mesh dimension
    med_mesh.attrs.create("ESP", dim)  # spatial dimension
    med_mesh.attrs.create("REP", 0)  # cartesian coordinate system (repère in French)
    med_mesh.attrs.create("UNT", numpy_void_str)  # time unit
    med_mesh.attrs.create("UNI", numpy_void_str)  # spatial unit
    med_mesh.attrs.create("SRT", 1)  # sorting type MED_SORT_ITDT

    # component names:
    names = ["X", "Y", "Z"][:dim]
    med_mesh.attrs.create("NOM", np.string_("".join(f"{name:<16}" for name in names)))
    med_mesh.attrs.create("DES", np.string_("Mesh created with adapy"))
    med_mesh.attrs.create("TYP", 0)  # mesh type (MED_NON_STRUCTURE)

    # Time-step
    step = "-0000000000000000001-0000000000000000001"  # NDT NOR
    time_step = med_mesh.create_group(step)
    time_step.attrs.create("CGT", 1)
    time_step.attrs.create("NDT", -1)  # no time step (-1)
    time_step.attrs.create("NOR", -1)  # no iteration step (-1)
    time_step.attrs.create("PDT", -1.0)  # current time
    return time_step



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.base.physical_objects import BackendGeom
from ada.base.types import BaseEnum
from ada.core.vector_utils import unit_vector

if TYPE_CHECKING:
    from ada import Beam, Node, Plate, PrimExtrude


class Bolts(BackendGeom):
    """

    TODO: Create a bolt class based on the IfcMechanicalFastener concept.

    https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/schema/ifcsharedcomponentelements/lexical/ifcmechanicalfastener.htm

    Which in turn should likely be inside another element components class

    https://standards.buildingsmart.org/IFC/RELEASE/IFC4_1/FINAL/HTML/schema/ifcsharedcomponentelements/lexical/ifcelementcomponent.htm

    """

    def __init__(self, name, p1, p2, normal, members, parent=None):
        super(Bolts, self).__init__(name, parent=parent)


class WeldProfileEnum(BaseEnum):
    V = "V"


class Weld(BackendGeom):
    def __init__(
        self,
        name,
        p1,
        p2,
        weld_type: WeldProfileEnum | str,
        members,
        profile: list[tuple],
        xdir: tuple,
        groove: list[tuple] = None,
        parent=None,
    ):
        super(Weld, self).__init__(name, parent=parent)
        from ada import Node, PrimExtrude

        p1 = Node(p1) if isinstance(p1, Node) is False else p1
        p2 = Node(p2) if isinstance(p2, Node) is False else p2
        vec = unit_vector(p2.p - p1.p)

        if isinstance(weld_type, str):
            weld_type = WeldProfileEnum.from_str(weld_type)

        if isinstance(profile, list):
            geom = PrimExtrude.from_2points_and_curve(f"{self.name}_geom", p1.p, p2.p, profile, xdir)
            geom.parent = self
        else:
            raise NotImplementedError()

        if groove is not None:
            p_start = p1.p - p1.p * vec * 0.02
            p_end = p2.p + p2.p * vec * 0.02
            groove = PrimExtrude.from_2points_and_curve(f"{self.name}_groove", p_start, p_end, groove, xdir)
            groove.parent = self

        self._xdir = xdir
        self._geom = geom
        self._groove = groove
        self._p1 = p1
        self._p2 = p2
        self._members = members
        self._weld_type = weld_type

    @property
    def type(self) -> WeldProfileEnum:
        return self._weld_type

    @property
    def p1(self) -> Node:
        return self._p1

    @property
    def p2(self) -> Node:
        return self._p2

    @property
    def members(self) -> list[Plate | Beam]:
        return self._members

    @property
    def geometry(self) -> PrimExtrude:
        return self._geom

    @property
    def groove(self) -> PrimExtrude:
        return self._groove



import os
import pathlib
from typing import Dict, List, Tuple, Union

from ada.config import get_logger

logger = get_logger()


def easy_plotly(
    title: str,
    in_data: Union[
        Tuple[list, list], List[tuple], Dict[str, Union[Tuple[float, float], Dict[str, Tuple[float, float]]]]
    ],
    xlbl: str = "X-axis",
    ylbl: str = "Y-axis",
    xrange: List[Union[float, int]] = None,
    yrange: List[Union[float, int]] = None,
    yaxformat: str = "E",
    xaxformat: str = None,
    legend_pos: Dict[str, float] = None,
    autoreverse=False,
    save_filename: Union[str, pathlib.PurePath, pathlib.Path] = None,
    mode="lines",
    marker="circle",
    traces=None,
    template="plotly_white",
    annotations=None,
    shapes=None,
    renderer="notebook_connected",
    return_widget=True,
    width=1600,
    height=800,
    log_x=False,
    log_y=False,
):
    """
    A Plotly template for quick and easy interactive scatter plotting using some pre-defined values. If you need more
    control of the plotly plot, you are probably better off using plotly directly

    See more information for scatter plots using Plotly at https://plot.ly/python/reference/#scatter

    :param title: Plot title
    :param in_data: tuple (x, y) for single plots or dict {'var1':{'x': [..], 'y': [..] }, 'var2': {..}, etc..}
    :param xlbl: X-axis label
    :param ylbl: Y-axis label
    :param xrange: min and max values of x-axis
    :param yrange: min and max values of y-axis
    :param yaxformat: "none" | "e" | "E" | "power" | "SI" | "B" (default) exponent format of y-axis.
    :param legend_pos: dict(x=-.1, y=1.2)
    :param autoreverse: Autoreverse the X-axis (opposed to inputting the reversed x-list)
    :param save_filename: Abs path to file location or file name of figure.
    :param mode:
    :param marker:
    :param traces: Add plotly traces manually
    :param template: Which plot template. Default is 'plotly_white'. Alternatives are shown below
    :param annotations:
    :param renderer: Which renderer should be used. Default is 'notebook_connected'. See below for alternatives
    :param return_widget:


    Templates:
                'ggplot2', 'seaborn', 'plotly', 'plotly_white', 'plotly_dark', 'presentation', 'xgridoff', 'none'

    renderers:
                'plotly_mimetype', 'jupyterlab', 'nteract', 'vscode', 'notebook', 'notebook_connected', 'kaggle',
                'azure', 'colab', 'cocalc', 'databricks', 'json', 'png', 'jpeg', 'jpg', 'svg', 'pdf', 'browser',
                'firefox', 'chrome', 'chromium', 'iframe', 'iframe_connected', 'sphinx_gallery'

    """
    import plotly.graph_objects as go

    plot_data = extract_plot_data(in_data, mode, marker)

    if traces is not None:
        plot_data += traces

    autorange = "reversed" if autoreverse is True else None
    layout = go.Layout(
        title=title,
        xaxis=dict(
            title=xlbl,
            titlefont=dict(family="Arial, monospace", size=18, color="#7f7f7f"),
            autorange=autorange,
            range=xrange,
            exponentformat=xaxformat,
        ),
        yaxis=dict(
            title=ylbl,
            titlefont=dict(family="Arial, monospace", size=18, color="#7f7f7f"),
            range=yrange,
            exponentformat=yaxformat,
        ),
        legend=legend_pos,
        template=template,
        shapes=shapes,
    )
    if annotations is not None:
        layout["annotations"] = annotations
    try:
        fig = go.FigureWidget(data=plot_data, layout=layout)
    except ImportError as e:
        fig = go.Figure(data=plot_data, layout=layout)
        logger.warning(f"Could not import go.FigureWidget due to ({e}).\nUsing go.Figure instead")

    if log_y is True:
        fig.update_yaxes(type="log", range=yrange, overwrite=True)  # log range: 10^0=1, 10^5=100000
    if log_x is True:
        fig.update_xaxes(type="log", range=xrange, overwrite=True)  # log range: 10^0=1, 10^5=100000

    if save_filename is not None:
        save_plot(fig, save_filename, width, height)
    else:
        if return_widget is True:
            return fig
        fig.show(renderer=renderer)


def save_plot(fig, save_filename, width, height):
    from plotly import io as pio

    filepath = pathlib.Path(save_filename)
    if filepath.suffix == "":
        filepath = filepath.with_suffix(".png")

    dirpath = os.path.dirname(filepath)
    print(f'Saving "{filepath}"')
    if os.path.isdir(dirpath) is False:
        os.makedirs(dirpath)
    pio.write_image(fig, filepath, width=width, height=height)


def extract_plot_data(in_data, mode, marker):
    import plotly.graph_objs as go

    plot_data = []
    if type(in_data) is dict:
        for key in in_data.keys():
            if type(in_data[key]) is dict:
                x_ = in_data[key]["x"]
                y_ = in_data[key]["y"]
            elif type(in_data[key]) in (tuple, list):
                x_ = in_data[key][0]
                y_ = in_data[key][1]
            else:
                raise Exception('Unrecognized input in dict "{}"'.format(type(in_data[key])))

            trace = go.Scatter(
                x=x_,
                y=y_,
                name=key,
                mode=mode,
                marker=dict(symbol=marker),
            )
            plot_data.append(trace)
    elif type(in_data) in [list, tuple]:
        if len(in_data) == 2:
            x, y = in_data
        else:
            x, y = zip(*in_data)
        trace = go.Scatter(
            x=x,
            y=y,
            mode=mode,
            marker=dict(symbol=marker),
        )
        plot_data.append(trace)
    else:
        raise Exception(f'Unrecognized input type "{type(in_data)}" found for "in_data" or "traces"')

    return plot_data



from typing import TYPE_CHECKING

from ada import Beam, Plate, Section
from ada.fem.meshing.partitioning.partition_beams import ibeam

if TYPE_CHECKING:
    from ada.fem.meshing.concepts import GmshData, GmshSession


def partition_objects_with_holes(gmsh_data: "GmshData", gmsh_session: "GmshSession"):
    obj = gmsh_data.obj

    partition_map = {Plate: partition_plate_with_hole}
    partition_tool = partition_map.get(type(obj), None)

    if partition_tool is None:
        raise NotImplementedError(f'Partitioning of "{type(obj)}" is not yet supported')

    partition_tool(gmsh_data, gmsh_session)


def partition_solid_beams(gmsh_data: "GmshData", gmsh_session: "GmshSession"):
    obj: Beam = gmsh_data.obj

    partition_map = {Section.TYPES.IPROFILE: ibeam}

    partition_tool = partition_map.get(obj.section.type, None)

    if partition_tool is None:
        raise NotImplementedError(f'Partitioning of "{obj.section.type}" Beams is not yet supported')

    partition_tool(gmsh_data, gmsh_session)


def partition_plate_with_hole(model: "GmshData", gmsh_session: "GmshSession"):
    gmsh_session.model.mesh.recombine()
    for dim, tag in model.entities:
        # ents.append(tag)
        # self.model.mesh.set_transfinite_surface(tag)
        gmsh_session.model.mesh.setRecombine(dim, tag)

    # gmsh_session.open_gui()
    #
    # raise NotImplementedError()



from typing import TYPE_CHECKING, Iterable

from ada.core.utils import NewLine
from ada.fem import Mass

from ..read.read_masses import ada_to_aba_mass_map
from .helper_utils import get_instance_name

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem import FemSet


def masses_str(fem: "FEM", written_on_assembly_level: bool):
    if len(list(fem.elements.masses)) == 0:
        return "** No Masses"

    return "\n".join([mass_str(m, written_on_assembly_level) for m in fem.elements.masses])


def mass_str(mass: Mass, written_on_assembly_level: bool) -> str:
    if mass.point_mass_type in (Mass.PTYPES.ISOTROPIC, None):
        type_str = ""
    else:
        aba_type = ada_to_aba_mass_map.get(mass.point_mass_type, None)
        if aba_type is None:
            raise NotImplementedError()
        type_str = f", type={aba_type}"

    mstr = ",".join([str(x) for x in mass.mass]) if type(mass.mass) is list else str(mass.mass)

    if mass.elset is not None:
        set_ref = mass.elset
    elif mass.fem_set is not None:
        set_ref = mass.fem_set
    else:
        raise ValueError("Unable to find proper reference to masses")
    set_name = get_instance_name(set_ref, written_on_assembly_level=written_on_assembly_level)
    if mass.type == Mass.TYPES.MASS:
        return f"""*Mass, elset={set_name}{type_str}\n {mstr}"""
    elif mass.type == Mass.TYPES.NONSTRU:
        return f"""*Nonstructural Mass, elset={set_name}, units={mass.units}\n  {mstr}"""
    elif mass.type == Mass.TYPES.ROT_INERTIA:
        return f"""*Rotary Inertia, elset={set_name}\n  {mstr}"""
    else:
        raise ValueError(f'Mass type "{mass.type}" is not supported by Abaqus')


def write_mass_elem(eltype: str, elset: "FemSet", fem: "FEM", elements: Iterable[Mass], alevel: bool) -> str:
    el_type = fem.options.ABAQUS.default_elements.get_element_type(eltype)
    el_set_str = f", ELSET={elset.name}" if elset is not None else ""
    if elset is None:
        return "** Masses not assigned to element sets\n"
    el_str = "\n".join((write_mass(el, alevel) for el in elements))
    return f"""*ELEMENT, type={el_type}{el_set_str}\n{el_str}\n"""


def write_mass(el: "Mass", alevel: bool) -> str:
    if el.nodes is None:
        return ""

    nl = NewLine(10, suffix=7 * " ")

    if len(el.nodes) > 6:
        di = " {}"
    else:
        di = "{:>13}"
    el_str = (
        f"{el.id:>7}, " + " ".join([f"{di.format(get_instance_name(no, alevel))}," + next(nl) for no in el.nodes])[:-1]
    )
    return el_str



from __future__ import annotations

from enum import Enum


class InvalidUnit(Exception):
    pass


class Units(Enum):
    M = "m"
    MM = "mm"

    @staticmethod
    def is_valid_unit(unit: str):
        return unit.lower() in list([x.value.lower() for x in Units])

    @staticmethod
    def from_str(unit: str) -> str:
        units_map = {x.value.lower(): x for x in Units}
        unit_safe = units_map.get(unit.lower())
        if unit_safe is None:
            raise InvalidUnit
        return unit_safe

    @staticmethod
    def get_scale_factor(from_unit, to_unit) -> float:
        if isinstance(from_unit, str):
            from_unit = Units.from_str(from_unit)
        if isinstance(to_unit, str):
            to_unit = Units.from_str(to_unit)

        scale_map = {
            (Units.MM, Units.M): 0.001,
            (Units.M, Units.M): 1.0,
            (Units.MM, Units.MM): 1.0,
            (Units.M, Units.MM): 1000.0,
        }
        result = scale_map.get((from_unit, to_unit))
        if result is None:
            raise InvalidUnit(f"Unable to convert {from_unit=} {to_unit=}")
        return result

    @staticmethod
    def get_general_point_tol(units: str | Units):
        from ada.config import Settings

        if isinstance(units, str):
            units = Units.from_str(units)

        if units == Units.MM:
            tol = Settings.mmtol
        elif units == Units.M:
            tol = Settings.mtol
        else:
            raise ValueError(f'Unknown unit "{units}"')
        return tol



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar

import numpy as np


@dataclass
class FieldData:
    name: str
    step: int
    components: list[str]
    values: np.ndarray
    eigen_freq: float = None
    eigen_value: float = None


class NodalFieldType(Enum):
    DISP = "displacement"
    VEL = "velocity"
    FORCE = "force"


@dataclass
class NodalFieldData(FieldData):
    COLS: ClassVar[list[str]] = ["node_label"]
    field_type: NodalFieldType = None

    def get_all_values(self):
        num_cols = len(self.COLS)
        num_comp = len(self.components)
        cr = list(range(num_cols, num_cols + num_comp + 1)) if num_comp != 0 else -1

        if num_comp == 0:
            return self.values[:, -1]

        result = self.values[:, cr[0] : cr[-1]]
        return result


class FieldPosition(Enum):
    NODAL = "nodal"
    INT = "integration_point"


@dataclass
class ElementFieldData(FieldData):
    """Values from element integration points"""

    field_pos: FieldPosition = FieldPosition.NODAL
    COLS: ClassVar[list[str]] = ["elem_label", "sec_num"]
    int_positions: list[tuple] = None

    def get_by_element_id(self, elem_ids: list[int], int_points: list[int] = None) -> ElementFieldData:
        data = []
        for x in self.values:
            if int(x[0]) not in elem_ids:
                continue
            if int_points is not None and x[1] not in int_points:
                continue
            data.append(x)

        return ElementFieldData(self.name, self.step, self.components, np.array(data))

    def _get_field_nodal(self, cr: int):
        nodes = self.values[:, 2]
        unique_nodes = np.unique(nodes)
        num_unique = len(unique_nodes)

        result_data = np.arange(0, num_unique)

        for i, unique_nodal in enumerate(unique_nodes):
            instances = np.where(nodes == unique_nodal)
            if cr != -1:
                field_data = self.values[instances, cr[0] : cr[-1]].ravel()
            else:
                field_data = self.values[instances, cr].ravel()
            averaged_data = sum(field_data) / len(field_data)
            result_data[i] = averaged_data

        return result_data

    def _get_field_int(self, cr) -> np.ndarray | dict[str, np.ndarray]:
        elements = self.values[:, 0]
        unique_element_ids = np.unique(elements)
        num_unique = len(unique_element_ids)

        if cr != -1:
            num_int_points = int(len(self.values) / num_unique)
            result_dict = {}
            for j, x in enumerate(self.components, start=cr[0]):
                result_data = np.zeros((num_unique, num_int_points))
                for i, unique_elements in enumerate(unique_element_ids):
                    instances = np.where(elements == unique_elements)
                    comp = self.values[instances, j]
                    result_data[i] = comp

                result_dict[x] = [result_data]

            return result_dict
        else:
            num_int_points = int(len(self.values) / num_unique) * len(self.components)
            result_data = np.zeros((num_unique, num_int_points))
            for i, unique_elements in enumerate(unique_element_ids):
                instances = np.where(elements == unique_elements)
                field_data = self.values[instances, cr].ravel()
                result_data[i] = field_data

                return result_data

    def get_all_values(self):
        num_cols = len(self.COLS)
        num_comp = len(self.components)
        cr = list(range(num_cols, num_cols + num_comp + 1)) if num_comp != 0 else -1

        if self.field_pos == self.field_pos.NODAL:
            return self._get_field_nodal(cr)
        else:
            return self._get_field_int(cr)


class LineSectionIntegrationPoints:
    ISection: dict = {1: "bottom left", 2: "bottom right", 3: "top left", 4: "top right"}



from itertools import groupby
from operator import itemgetter
from typing import TYPE_CHECKING

from ada.fem import FemSet
from ada.fem.containers import FemSets
from ada.fem.formats.utils import str_to_int

from . import cards

if TYPE_CHECKING:
    from ada import FEM


def get_sets(bulk_str: str, parent: "FEM") -> FemSets:
    set_map = dict()
    set_groups = (get_setmap(m, parent) for m in cards.re_setmembs.finditer(bulk_str))

    for setid_el_type, content in groupby(set_groups, key=itemgetter(0, 1)):
        setid = setid_el_type[0]
        eltype = setid_el_type[1]
        set_map[setid] = [list(), eltype]
        for c in content:
            set_map[setid][0] += c[2]
    fem_sets = [get_femsets(m, set_map, parent) for m in cards.re_setnames.finditer(bulk_str)]
    return FemSets(fem_sets, parent=parent)


def get_setmap(m, parent):
    d = m.groupdict()
    set_type = "nset" if str_to_int(d["istype"]) == 1 else "elset"
    mem_list = d["members"].split()
    if set_type == "nset":
        members = [parent.nodes.from_id(str_to_int(x)) for x in mem_list]
    else:
        members = [parent.elements.from_id(str_to_int(x)) for x in mem_list]
    return str_to_int(d["isref"]), set_type, members


def get_femsets(m, set_map, parent) -> FemSet:
    d = m.groupdict()
    isref = str_to_int(d["isref"])
    fem_set = FemSet(
        d["set_name"].strip(),
        set_map[isref][0],
        set_map[isref][1],
        parent=parent,
    )
    return fem_set



from __future__ import annotations

from ada.fem.shapes.definitions import (
    LineShapes,
    MassTypes,
    ShellShapes,
    SolidShapes,
    SpringTypes,
)

sesam_el_map = {
    15: LineShapes.LINE,
    2: LineShapes.LINE,
    23: LineShapes.LINE3,
    24: ShellShapes.QUAD,
    25: ShellShapes.TRI,
    26: ShellShapes.TRI6,
    28: ShellShapes.QUAD8,
    31: SolidShapes.TETRA10,
    40: SpringTypes.SPRING2,
    18: SpringTypes.SPRING1,
    11: MassTypes.MASS,
}

sesam_reverse = {value: key for key, value in sesam_el_map.items()}


def sesam_eltype_2_general(eltyp: int) -> LineShapes | ShellShapes | SolidShapes | SpringTypes | MassTypes:
    """Converts the numeric definition of elements in Sesam to a generalized element type form (ie. B31, S4, etc..)"""
    res = sesam_el_map.get(eltyp, None)
    if res is None:
        raise Exception("Currently unsupported eltype", eltyp)
    return res



from __future__ import annotations

from ada.concepts.primitives import (
    BSplineSurfaceWithKnots,
    IfcBSplineSurfaceForm,
    RationalBSplineSurfaceWithKnots,
)
from ada.config import get_logger

logger = get_logger()


def create_bsplinesurface_from_sat(spline_data_str: str) -> BSplineSurfaceWithKnots | RationalBSplineSurfaceWithKnots:
    head, data = spline_data_str.split("{")

    data_lines = [x.strip() for x in data.splitlines()]
    dline = data_lines[0].split()
    u_degree, v_degree = [int(float(x)) for x in dline[3:5]]
    surf_form = IfcBSplineSurfaceForm.UNSPECIFIED
    uknots_in = [float(x) for x in data_lines[1].split()]
    uknots = uknots_in[0::2]
    uMult = [int(x) for x in uknots_in[1::2]]

    vknots_in = [float(x) for x in data_lines[2].split()]
    vknots = vknots_in[0::2]

    vMult = [int(x) for x in vknots_in[1::2]]
    res = [[float(i) for i in x.split()] for x in data_lines[3 : 3 + v_degree * 4]]
    control_points = []
    for i in range(0, v_degree):
        control_points += [res[i::3]]

    weights = None
    if len(control_points[0]) == 4:
        weights = [[i[-1] for i in x] for x in control_points]

    if dline[0] == "exactsur":
        logger.info("Exact surface")

    props = dict(
        uDegree=u_degree,
        vDegree=v_degree,
        controlPointsList=control_points,
        surfaceForm=surf_form,
        uKnots=uknots,
        vKnots=vknots,
        uMultiplicities=uMult,
        vMultiplicities=vMult,
    )

    if weights is not None:
        surface = RationalBSplineSurfaceWithKnots(**props, weightsData=weights)
    else:
        surface = BSplineSurfaceWithKnots(**props)

    return surface
